"""
TornadoService is the base class for your handlers.
It directly inherits from :py:class:`tornado.web.RequestHandler`
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from io import open

import os
import time
import threading
from datetime import datetime
from six.moves import http_client
from tornado.web import RequestHandler, HTTPError
from tornado import gen
import tornado.ioloop
from tornado.ioloop import IOLoop

import DIRAC

from DIRAC import gConfig, gLogger, S_OK
from DIRAC.ConfigurationSystem.Client import PathFinder
from DIRAC.Core.DISET.AuthManager import AuthManager
from DIRAC.Core.Security.X509Chain import X509Chain  # pylint: disable=import-error
from DIRAC.Core.Utilities.JEncode import decode, encode
from DIRAC.FrameworkSystem.Client.MonitoringClient import MonitoringClient

sLog = gLogger.getSubLogger(__name__)


class TornadoService(RequestHandler):  # pylint: disable=abstract-method
    """
    Base class for all the Handlers.
    It directly inherits from :py:class:`tornado.web.RequestHandler`

    Each HTTP request is served by a new instance of this class.

    For the sequence of method called, please refer to
    the `tornado documentation <https://www.tornadoweb.org/en/stable/guide/structure.html>`_.

    For compatibility with the existing :py:class:`DIRAC.Core.DISET.TransferClient.TransferClient`,
    the handler can define a method ``export_streamToClient``. This is the method that will be called
    whenever ``TransferClient.receiveFile`` is called. It is the equivalent of the DISET
    ``transfer_toClient``.
    Note that this is here only for compatibility, and we discourage using it for new purposes, as it is
    bound to disappear.

    The handler only define the ``post`` verb. Please refer to :py:meth:`.post` for the details.

    In order to create a handler for your service, it has to
    follow a certain skeleton::

      from DIRAC.Core.Tornado.Server.TornadoService import TornadoService
      class yourServiceHandler(TornadoService):

        # Called only once when the first
        # request for this handler arrives
        # Useful for initializing DB or so.
        # You don't need to use super or to call any parents method, it's managed by the server
        @classmethod
        def initializeHandler(cls, infosDict):
          '''Called only once when the first
             request for this handler arrives
             Useful for initializing DB or so.
             You don't need to use super or to call any parents method, it's managed by the server
          '''
          pass


        def initializeRequest(self):
          '''
             Called at the beginning of each request
          '''
          pass

        # Specify the default permission for the method
        # See :py:class:`DIRAC.Core.DISET.AuthManager.AuthManager`
        auth_someMethod = ['authenticated']


        def export_someMethod(self):
          '''The method you want to export.
           It must start with ``export_``
           and it must return an S_OK/S_ERROR structure
          '''
          return S_ERROR()


        def export_streamToClient(self, myDataToSend, token):
          ''' Automatically called when ``Transfer.receiveFile`` is called.
              Contrary to the other ``export_`` methods, it does not need
              to return a DIRAC structure.
          '''

          # Do whatever with the token

          with open(myFileToSend, 'r') as fd:
            return fd.read()


    Note that because we inherit from :py:class:`tornado.web.RequestHandler`
    and we are running using executors, the methods you export cannot write
    back directly to the client. Please see inline comments for more details.

    In order to pass information around and keep some states, we use instance attributes.
    These are initialized in the :py:meth:`.initialize` method.

    """

    # Because we initialize at first request, we use a flag to know if it's already done
    __init_done = False
    # Lock to make sure that two threads are not initializing at the same time
    __init_lock = threading.RLock()

    # MonitoringClient, we don't use gMonitor which is not thread-safe
    # We also need to add specific attributes for each service
    _monitor = None

    @classmethod
    def _initMonitoring(cls, serviceName, fullUrl):
        """
        Initialize the monitoring specific to this handler
        This has to be called only by :py:meth:`.__initializeService`
        to ensure thread safety and unicity of the call.

        :param serviceName: relative URL ``/<System>/<Component>``
        :param fullUrl: full URl like ``https://<host>:<port>/<System>/<Component>``
        """

        # Init extra bits of monitoring

        cls._monitor = MonitoringClient()
        cls._monitor.setComponentType(MonitoringClient.COMPONENT_WEB)

        cls._monitor.initialize()

        if tornado.process.task_id() is None:  # Single process mode
            cls._monitor.setComponentName("Tornado/%s" % serviceName)
        else:
            cls._monitor.setComponentName("Tornado/CPU%d/%s" % (tornado.process.task_id(), serviceName))

        cls._monitor.setComponentLocation(fullUrl)

        cls._monitor.registerActivity("Queries", "Queries served", "Framework", "queries", MonitoringClient.OP_RATE)

        cls._monitor.setComponentExtraParam("DIRACVersion", DIRAC.version)
        cls._monitor.setComponentExtraParam("platform", DIRAC.getPlatform())
        cls._monitor.setComponentExtraParam("startTime", datetime.utcnow())

        cls._stats = {"requests": 0, "monitorLastStatsUpdate": time.time()}

        return S_OK()

    @classmethod
    def __initializeService(cls, relativeUrl, absoluteUrl):
        """
        Initialize a service.
        The work is only perform once at the first request.

        :param relativeUrl: relative URL, e.g. ``/<System>/<Component>``
        :param absoluteUrl: full URL e.g. ``https://<host>:<port>/<System>/<Component>``

        :returns: S_OK
        """
        # If the initialization was already done successfuly,
        # we can just return
        if cls.__init_done:
            return S_OK()

        # Otherwise, do the work but with a lock
        with cls.__init_lock:

            # Check again that the initialization was not done by another thread
            # while we were waiting for the lock
            if cls.__init_done:
                return S_OK()

            # Url starts with a "/", we just remove it
            serviceName = relativeUrl[1:]

            cls._startTime = datetime.utcnow()
            sLog.info("First use, initializing service...", "%s" % relativeUrl)
            cls._authManager = AuthManager("%s/Authorization" % PathFinder.getServiceSection(serviceName))

            cls._initMonitoring(serviceName, absoluteUrl)

            cls._serviceName = serviceName
            cls._validNames = [serviceName]
            serviceInfo = {
                "serviceName": serviceName,
                "serviceSectionPath": PathFinder.getServiceSection(serviceName),
                "csPaths": [PathFinder.getServiceSection(serviceName)],
                "URL": absoluteUrl,
            }
            cls._serviceInfoDict = serviceInfo

            cls.__monitorLastStatsUpdate = time.time()

            cls.initializeHandler(serviceInfo)

            cls.__init_done = True

            return S_OK()

    @classmethod
    def initializeHandler(cls, serviceInfoDict):
        """
        This may be overwritten when you write a DIRAC service handler
        And it must be a class method. This method is called only one time,
        at the first request

        :param dict ServiceInfoDict: infos about services, it contains
                                      'serviceName', 'serviceSectionPath',
                                      'csPaths' and 'URL'
        """
        pass

    def initializeRequest(self):
        """
        Called at every request, may be overwritten in your handler.
        """
        pass

    # This is a Tornado magic method
    def initialize(self):  # pylint: disable=arguments-differ
        """
        Initialize the handler, called at every request.

        It just calls :py:meth:`.__initializeService`

        If anything goes wrong, the client will get ``Connection aborted``
        error. See details inside the method.

        ..warning::
          DO NOT REWRITE THIS FUNCTION IN YOUR HANDLER
          ==> initialize in DISET became initializeRequest in HTTPS !
        """

        # Only initialized once
        if not self.__init_done:
            # Ideally, if something goes wrong, we would like to return a Server Error 500
            # but this method cannot write back to the client as per the
            # `tornado doc <https://www.tornadoweb.org/en/stable/guide/structure.html#overriding-requesthandler-methods>`_.
            # So the client will get a ``Connection aborted```
            try:
                res = self.__initializeService(self.srv_getURL(), self.request.full_url())
                if not res["OK"]:
                    raise Exception(res["Message"])
            except Exception as e:
                sLog.error("Error in initialization", repr(e))
                raise

    def prepare(self):
        """
        Prepare the request. It reads certificates and check authorizations.
        We make the assumption that there is always going to be a ``method`` argument
        regardless of the HTTP method used

        """

        # "method" argument of the POST call.
        # This resolves into the ``export_<method>`` method
        # on the handler side
        # If the argument is not available, the method exists
        # and an error 400 ``Bad Request`` is returned to the client
        self.method = self.get_argument("method")

        self._stats["requests"] += 1
        self._monitor.setComponentExtraParam("queries", self._stats["requests"])
        self._monitor.addMark("Queries")

        try:
            self.credDict = self._gatherPeerCredentials()
        except Exception:  # pylint: disable=broad-except
            # If an error occur when reading certificates we close connection
            # It can be strange but the RFC, for HTTP, say's that when error happend
            # before authentication we return 401 UNAUTHORIZED instead of 403 FORBIDDEN
            sLog.error("Error gathering credentials", "%s; path %s" % (self.getRemoteAddress(), self.request.path))
            raise HTTPError(status_code=http_client.UNAUTHORIZED)

        # Resolves the hard coded authorization requirements
        try:
            hardcodedAuth = getattr(self, "auth_" + self.method)
        except AttributeError:
            hardcodedAuth = None

        # Check whether we are authorized to perform the query
        # Note that performing the authQuery modifies the credDict...
        authorized = self._authManager.authQuery(self.method, self.credDict, hardcodedAuth)
        if not authorized:
            sLog.error(
                "Unauthorized access",
                "Identity %s; path %s; DN %s"
                % (
                    self.srv_getFormattedRemoteCredentials,
                    self.request.path,
                    self.credDict["DN"],
                ),
            )
            raise HTTPError(status_code=http_client.UNAUTHORIZED)

    # Make post a coroutine.
    # See https://www.tornadoweb.org/en/branch5.1/guide/coroutines.html#coroutines
    # for details
    @gen.coroutine
    def post(self):  # pylint: disable=arguments-differ
        """
        Method to handle incoming ``POST`` requests.
        Note that all the arguments are already prepared in the :py:meth:`.prepare`
        method.

        The ``POST`` arguments expected are:

        * ``method``: name of the method to call
        * ``args``: JSON encoded arguments for the method
        * ``extraCredentials``: (optional) Extra informations to authenticate client
        * ``rawContent``: (optionnal, default False) If set to True, return the raw output
          of the method called.

        If ``rawContent`` was requested by the client, the ``Content-Type``
        is ``application/octet-stream``, otherwise we set it to ``application/json``
        and JEncode retVal.

        If ``retVal`` is a dictionary that contains a ``Callstack`` item,
        it is removed, not to leak internal information.


        Example of call using ``requests``::

          In [20]: url = 'https://server:8443/DataManagement/TornadoFileCatalog'
            ...: cert = '/tmp/x509up_u1000'
            ...: kwargs = {'method':'whoami'}
            ...: caPath = '/home/dirac/ClientInstallDIR/etc/grid-security/certificates/'
            ...: with requests.post(url, data=kwargs, cert=cert, verify=caPath) as r:
            ...:     print r.json()
            ...:
          {u'OK': True,
              u'Value': {u'DN': u'/C=ch/O=DIRAC/OU=DIRAC CI/CN=ciuser',
              u'group': u'dirac_user',
              u'identity': u'/C=ch/O=DIRAC/OU=DIRAC CI/CN=ciuser',
              u'isLimitedProxy': False,
              u'isProxy': True,
              u'issuer': u'/C=ch/O=DIRAC/OU=DIRAC CI/CN=ciuser',
              u'properties': [u'NormalUser'],
              u'secondsLeft': 85441,
              u'subject': u'/C=ch/O=DIRAC/OU=DIRAC CI/CN=ciuser/CN=2409820262',
              u'username': u'adminusername',
              u'validDN': False,
              u'validGroup': False}}
        """

        sLog.notice(
            "Incoming request",
            "%s /%s: %s" % (self.srv_getFormattedRemoteCredentials(), self._serviceName, self.method),
        )

        # Execute the method in an executor (basically a separate thread)
        # Because of that, we cannot calls certain methods like `self.write`
        # in __executeMethod. This is because these methods are not threadsafe
        # https://www.tornadoweb.org/en/branch5.1/web.html#thread-safety-notes
        # However, we can still rely on instance attributes to store what should
        # be sent back (reminder: there is an instance
        # of this class created for each request)
        retVal = yield IOLoop.current().run_in_executor(None, self.__executeMethod)

        # retVal is :py:class:`tornado.concurrent.Future`
        self.result = retVal.result()

        # Strip the exception/callstack info from S_ERROR responses
        if isinstance(self.result, dict):
            # ExecInfo comes from the exception
            if "ExecInfo" in self.result:
                del self.result["ExecInfo"]
            # CallStack comes from the S_ERROR construction
            if "CallStack" in self.result:
                del self.result["CallStack"]

        # Here it is safe to write back to the client, because we are not
        # in a thread anymore

        # If set to true, do not JEncode the return of the RPC call
        # This is basically only used for file download through
        # the 'streamToClient' method.
        rawContent = self.get_argument("rawContent", default=False)

        if rawContent:
            # See 4.5.1 http://www.rfc-editor.org/rfc/rfc2046.txt
            self.set_header("Content-Type", "application/octet-stream")
            result = self.result
        else:
            self.set_header("Content-Type", "application/json")
            result = encode(self.result)

        self.write(result)
        self.finish()

    # This nice idea of streaming to the client cannot work because we are ran in an executor
    # and we should not write back to the client in a different thread.
    # See https://www.tornadoweb.org/en/branch5.1/web.html#thread-safety-notes
    # def export_streamToClient(self, filename):
    #   # https://bhch.github.io/posts/2017/12/serving-large-files-with-tornado-safely-without-blocking/
    #   #import ipdb; ipdb.set_trace()
    #   # chunk size to read
    #   chunk_size = 1024 * 1024 * 1  # 1 MiB

    #   with open(filename, 'rb') as f:
    #     while True:
    #       chunk = f.read(chunk_size)
    #       if not chunk:
    #         break
    #       try:
    #         self.write(chunk)  # write the chunk to response
    #         self.flush()  # send the chunk to client
    #       except StreamClosedError:
    #         # this means the client has closed the connection
    #         # so break the loop
    #         break
    #       finally:
    #         # deleting the chunk is very important because
    #         # if many clients are downloading files at the
    #         # same time, the chunks in memory will keep
    #         # increasing and will eat up the RAM
    #         del chunk
    #         # pause the coroutine so other handlers can run
    #         yield gen.sleep(0.000000001)  # 1 nanosecond

    #   return S_OK()

    @gen.coroutine
    def __executeMethod(self):
        """
        Execute the method called, this method is ran in an executor
        We have several try except to catch the different problem which can occur

        - First, the method does not exist => Attribute error, return an error to client
        - second, anything happend during execution => General Exception, send error to client

        .. warning::
          This method is called in an executor, and so cannot use methods like self.write
          See https://www.tornadoweb.org/en/branch5.1/web.html#thread-safety-notes
        """
        # getting method
        try:
            # For compatibility reasons with DISET, the methods are still called ``export_*``
            method = getattr(self, "export_%s" % self.method)
        except AttributeError:
            sLog.error("Invalid method", self.method)
            raise HTTPError(status_code=http_client.NOT_IMPLEMENTED)

        # Decode args
        args_encoded = self.get_body_argument("args", default=encode([]))

        args = decode(args_encoded)[0]
        # Execute
        try:
            self.initializeRequest()
            retVal = method(*args)
        except Exception as e:  # pylint: disable=broad-except
            sLog.exception("Exception serving request", "%s:%s" % (str(e), repr(e)))
            raise HTTPError(http_client.INTERNAL_SERVER_ERROR)

        return retVal

    # def __write_return(self, retVal):
    #   """
    #     Write back to the client and return.
    #     It sets some headers (status code, ``Content-Type``).
    #     If raw content was requested by the client, the ``Content-Type``
    #     is ``application/octet-stream``, otherwise we set it to ``application/json``
    #     and JEncode retVal.

    #     If ``retVal`` is a dictionary that contains a ``Callstack`` item,
    #     it is removed, not to leak internal information.

    #     :param retVal: anything that can be serialized in json.
    #   """

    #   # In case of error in server side we hide server CallStack to client
    #   try:
    #     if 'CallStack' in retVal:
    #       del retVal['CallStack']
    #   except TypeError:
    #     pass

    #   # Set the status
    #   self.set_status(self._httpStatus)

    #   # This is basically only used for file download through
    #   # the 'streamToClient' method.
    #   if self.rawContent:
    #     # See 4.5.1 http://www.rfc-editor.org/rfc/rfc2046.txt
    #     self.set_header("Content-Type", "application/octet-stream")
    #     returnedData = retVal
    #   else:
    #     self.set_header("Content-Type", "application/json")
    #     returnedData = encode(retVal)

    #   self.write(returnedData)

    def on_finish(self):
        """
        Called after the end of HTTP request.
        Log the request duration
        """
        elapsedTime = 1000.0 * self.request.request_time()

        try:
            if self.result["OK"]:
                argsString = "OK"
            else:
                argsString = "ERROR: %s" % self.result["Message"]
        except (AttributeError, KeyError):  # In case it is not a DIRAC structure
            if self._reason == "OK":
                argsString = "OK"
            else:
                argsString = "ERROR %s" % self._reason

            argsString = "ERROR: %s" % self._reason
        sLog.notice(
            "Returning response",
            "%s %s (%.2f ms) %s"
            % (self.srv_getFormattedRemoteCredentials(), self._serviceName, elapsedTime, argsString),
        )

    def _gatherPeerCredentials(self):
        """
        Load client certchain in DIRAC and extract informations.

        The dictionary returned is designed to work with the AuthManager,
        already written for DISET and re-used for HTTPS.

        :returns: a dict containing the return of :py:meth:`DIRAC.Core.Security.X509Chain.X509Chain.getCredentials`
                  (not a DIRAC structure !)
        """

        chainAsText = self.request.get_ssl_certificate().as_pem()
        peerChain = X509Chain()

        # Here we read all certificate chain
        cert_chain = self.request.get_ssl_certificate_chain()
        for cert in cert_chain:
            chainAsText += cert.as_pem()

        peerChain.loadChainFromString(chainAsText)

        # Retrieve the credentials
        res = peerChain.getCredentials(withRegistryInfo=False)
        if not res["OK"]:
            raise Exception(res["Message"])

        credDict = res["Value"]

        # We check if client sends extra credentials...
        if "extraCredentials" in self.request.arguments:
            extraCred = self.get_argument("extraCredentials")
            if extraCred:
                credDict["extraCredentials"] = decode(extraCred)[0]
        return credDict

    ####
    #
    #   Default method
    #
    ####

    auth_ping = ["all"]

    def export_ping(self):
        """
        Default ping method, returns some info about server.

        It returns the exact same information as DISET, for transparency purpose.
        """
        # COPY FROM DIRAC.Core.DISET.RequestHandler
        dInfo = {}
        dInfo["version"] = DIRAC.version
        dInfo["time"] = datetime.utcnow()
        # Uptime
        try:
            with open("/proc/uptime", "rt") as oFD:
                iUptime = int(float(oFD.readline().split()[0].strip()))
            dInfo["host uptime"] = iUptime
        except Exception:  # pylint: disable=broad-except
            pass
        startTime = self._startTime
        dInfo["service start time"] = self._startTime
        serviceUptime = datetime.utcnow() - startTime
        dInfo["service uptime"] = serviceUptime.days * 3600 + serviceUptime.seconds
        # Load average
        try:
            with open("/proc/loadavg", "rt") as oFD:
                dInfo["load"] = " ".join(oFD.read().split()[:3])
        except Exception:  # pylint: disable=broad-except
            pass
        dInfo["name"] = self._serviceInfoDict["serviceName"]
        stTimes = os.times()
        dInfo["cpu times"] = {
            "user time": stTimes[0],
            "system time": stTimes[1],
            "children user time": stTimes[2],
            "children system time": stTimes[3],
            "elapsed real time": stTimes[4],
        }

        return S_OK(dInfo)

    auth_echo = ["all"]

    @staticmethod
    def export_echo(data):
        """
        This method used for testing the performance of a service
        """
        return S_OK(data)

    auth_whoami = ["authenticated"]

    def export_whoami(self):
        """
        A simple whoami, returns all credential dictionary, except certificate chain object.
        """
        credDict = self.srv_getRemoteCredentials()
        if "x509Chain" in credDict:
            # Not serializable
            del credDict["x509Chain"]
        return S_OK(credDict)

    ####
    #
    #  Utilities methods, some getters.
    #  From DIRAC.Core.DISET.requestHandler to get same interface in the handlers.
    #  Adapted for Tornado.
    #  These method are copied from DISET RequestHandler, they are not all used when i'm writing
    #  these lines. I rewrite them for Tornado to get them ready when a new HTTPS service need them
    #
    ####

    @classmethod
    def srv_getCSOption(cls, optionName, defaultValue=False):
        """
        Get an option from the CS section of the services

        :return: Value for serviceSection/optionName in the CS being defaultValue the default
        """
        if optionName[0] == "/":
            return gConfig.getValue(optionName, defaultValue)
        for csPath in cls._serviceInfoDict["csPaths"]:
            result = gConfig.getOption(
                "%s/%s"
                % (
                    csPath,
                    optionName,
                ),
                defaultValue,
            )
            if result["OK"]:
                return result["Value"]
        return defaultValue

    def getCSOption(self, optionName, defaultValue=False):
        """
        Just for keeping same public interface
        """
        return self.srv_getCSOption(optionName, defaultValue)

    def srv_getRemoteAddress(self):
        """
        Get the address of the remote peer.

        :return: Address of remote peer.
        """

        remote_ip = self.request.remote_ip
        # Although it would be trivial to add this attribute in _HTTPRequestContext,
        # Tornado won't release anymore 5.1 series, so go the hacky way
        try:
            remote_port = self.request.connection.stream.socket.getpeername()[1]
        except Exception:  # pylint: disable=broad-except
            remote_port = 0

        return (remote_ip, remote_port)

    def getRemoteAddress(self):
        """
        Just for keeping same public interface
        """
        return self.srv_getRemoteAddress()

    def srv_getRemoteCredentials(self):
        """
        Get the credentials of the remote peer.

        :return: Credentials dictionary of remote peer.
        """
        return self.credDict

    def getRemoteCredentials(self):
        """
        Get the credentials of the remote peer.

        :return: Credentials dictionary of remote peer.
        """
        return self.credDict

    def srv_getFormattedRemoteCredentials(self):
        """
        Return the DN of user

        Mostly copy paste from
        :py:meth:`DIRAC.Core.DISET.private.Transports.BaseTransport.BaseTransport.getFormattedCredentials`

        Note that the information will be complete only once the AuthManager was called
        """
        address = self.getRemoteAddress()
        peerId = ""
        # Depending on where this is call, it may be that credDict is not yet filled.
        # (reminder: AuthQuery fills part of it..)
        try:
            peerId = "[%s:%s]" % (self.credDict["group"], self.credDict["username"])
        except AttributeError:
            pass

        if address[0].find(":") > -1:
            return "([%s]:%s)%s" % (address[0], address[1], peerId)
        return "(%s:%s)%s" % (address[0], address[1], peerId)

    # def getFormattedCredentials(self):
    #     peerCreds = self.getConnectingCredentials()
    #     address = self.getRemoteAddress()
    #     if 'username' in peerCreds:
    #       peerId = "[%s:%s]" % (peerCreds['group'], peerCreds['username'])
    #     else:
    #       peerId = ""
    #     if address[0].find(":") > -1:
    #       return "([%s]:%s)%s" % (address[0], address[1], peerId)
    #     return "(%s:%s)%s" % (address[0], address[1], peerId)

    def srv_getServiceName(self):
        """
        Return the service name
        """
        return self._serviceInfoDict["serviceName"]

    def srv_getURL(self):
        """
        Return the URL
        """
        return self.request.path



########################################################################
# $HeadURL $
# File: ReTransfer.py
# Author: Krzysztof.Ciba@NOSPAMgmail.com
# Date: 2013/04/02 14:24:21
########################################################################
""" :mod: ReTransfer

    ================


    .. module: ReTransfer

    :synopsis: ReTransfer Operation handler

    .. moduleauthor:: Krzysztof.Ciba@NOSPAMgmail.com

    ReTransfer Operation handler
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id $"

# #
# @file ReTransfer.py
# @author Krzysztof.Ciba@NOSPAMgmail.com
# @date 2013/04/02 14:24:31
# @brief Definition of ReTransfer class.

# # imports
from DIRAC import S_OK, S_ERROR
from DIRAC.FrameworkSystem.Client.MonitoringClient import gMonitor
from DIRAC.DataManagementSystem.Agent.RequestOperations.DMSRequestOperationsBase import DMSRequestOperationsBase
from DIRAC.Resources.Storage.StorageElement import StorageElement

from DIRAC.MonitoringSystem.Client.MonitoringReporter import MonitoringReporter

########################################################################


class ReTransfer(DMSRequestOperationsBase):
    """
    .. class:: ReTransfer


    online ReTransfer operation handler

    :param self: self reference
    :param ~DIRAC.RequestManagementSystem.Client.Operation.Operation operation: Operation instance
    :param str csPath: CS path for this handler

    """

    def __init__(self, operation=None, csPath=None):
        """c'tor"""
        # # base class ctor
        DMSRequestOperationsBase.__init__(self, operation, csPath)

    def __call__(self):
        """reTransfer operation execution"""

        # The flag  'rmsMonitoring' is set by the RequestTask and is False by default.
        # Here we use 'createRMSRecord' to create the ES record which is defined inside OperationHandlerBase.
        if self.rmsMonitoring:
            self.rmsMonitoringReporter = MonitoringReporter(monitoringType="RMSMonitoring")
        else:
            # # gMonitor stuff
            gMonitor.registerActivity(
                "FileReTransferAtt", "File retransfers attempted", "RequestExecutingAgent", "Files/min", gMonitor.OP_SUM
            )
            gMonitor.registerActivity(
                "FileReTransferOK", "File retransfers successful", "RequestExecutingAgent", "Files/min", gMonitor.OP_SUM
            )
            gMonitor.registerActivity(
                "FileReTransferFail", "File retransfers failed", "RequestExecutingAgent", "Files/min", gMonitor.OP_SUM
            )

        # # list of targetSEs
        targetSEs = self.operation.targetSEList
        # # check targetSEs for removal
        targetSE = targetSEs[0]
        bannedTargets = self.checkSEsRSS(targetSE)
        if not bannedTargets["OK"]:
            if self.rmsMonitoring:
                for status in ["Attempted", "Failed"]:
                    self.rmsMonitoringReporter.addRecord(self.createRMSRecord(status, len(self.operation)))
                self.rmsMonitoringReporter.commit()
            else:
                gMonitor.addMark("FileReTransferAtt")
                gMonitor.addMark("FileReTransferFail")
            return bannedTargets

        if bannedTargets["Value"]:
            return S_OK("%s targets are banned for writing" % ",".join(bannedTargets["Value"]))

        # # get waiting files
        waitingFiles = self.getWaitingFilesList()
        # # prepare waiting files
        toRetransfer = dict([(opFile.PFN, opFile) for opFile in waitingFiles])

        if self.rmsMonitoring:
            self.rmsMonitoringReporter.addRecord(self.createRMSRecord("Attempted", len(toRetransfer)))
        else:
            gMonitor.addMark("FileReTransferAtt", len(toRetransfer))

        if len(targetSEs) != 1:
            error = "only one TargetSE allowed, got %d" % len(targetSEs)
            for opFile in toRetransfer.values():
                opFile.Error = error
                opFile.Status = "Failed"
            self.operation.Error = error

            if self.rmsMonitoring:
                self.rmsMonitoringReporter.addRecord(self.createRMSRecord("Failed", len(toRetransfer)))
                self.rmsMonitoringReporter.commit()
            else:
                gMonitor.addMark("FileReTransferFail", len(toRetransfer))

            return S_ERROR(error)

        se = StorageElement(targetSE)
        for opFile in toRetransfer.values():

            reTransfer = se.retransferOnlineFile(opFile.LFN)
            if not reTransfer["OK"]:
                opFile.Error = reTransfer["Message"]
                self.log.error("Retransfer failed", opFile.Error)

                if self.rmsMonitoring:
                    self.rmsMonitoringReporter.addRecord(self.createRMSRecord("Failed", 1))
                else:
                    gMonitor.addMark("FileReTransferFail", 1)

                continue
            reTransfer = reTransfer["Value"]
            if opFile.LFN in reTransfer["Failed"]:
                opFile.Error = reTransfer["Failed"][opFile.LFN]
                self.log.error("Retransfer failed", opFile.Error)

                if self.rmsMonitoring:
                    self.rmsMonitoringReporter.addRecord(self.createRMSRecord("Failed", 1))
                else:
                    gMonitor.addMark("FileReTransferFail", 1)

                continue
            opFile.Status = "Done"
            self.log.info("%s retransfer done" % opFile.LFN)

            if self.rmsMonitoring:
                self.rmsMonitoringReporter.addRecord(self.createRMSRecord("Successful", 1))
            else:
                gMonitor.addMark("FileReTransferOK", 1)

        if self.rmsMonitoring:
            self.rmsMonitoringReporter.commit()

        return S_OK()



""" DIRAC script """
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import functools
import os.path
import sys
import six

from DIRAC.Core.Utilities.DErrno import includeExtensionErrors
from DIRAC.FrameworkSystem.Client.Logger import gLogger
from DIRAC.FrameworkSystem.Client.MonitoringClient import gMonitor
from DIRAC.ConfigurationSystem.Client.LocalConfiguration import LocalConfiguration

if six.PY3:
    from DIRAC.Core.Utilities.Extensions import entrypointToExtension, extensionsByPriority


class DIRACScript(object):
    """Decorator for providing command line executables

    All console-scripts entrypoints in DIRAC and downstream extensions should be
    wrapped in this decorator to allow extensions to override any entry_point.
    """

    scriptName = None
    alreadyInitialized = False
    localCfg = LocalConfiguration()

    def __call__(self, func=None):
        """Set the wrapped function or call the script

        This function is either called with a decorator or directly to call the
        underlying function. When running with Python 2 the raw function will always
        be called however in Python 3 the priorities will be applied from the
        dirac.extension_metadata entry_point.
        """
        # If func is provided then the decorator is being applied to a function
        if func is not None:
            self._func = func
            # Find the name of the command and its documentation
            DIRACScript.localCfg.setUsageMessage(func.__globals__["__doc__"])
            DIRACScript.scriptName = os.path.basename(func.__globals__["__file__"])[:-3].replace("_", "-")
            return functools.wraps(func)(self)

        # Setuptools based installations aren't supported with Python 2
        if six.PY2:
            return self._func()  # pylint: disable=not-callable

        # This is only available in Python 3.8+ so it has to be here for now
        from importlib import metadata  # pylint: disable=no-name-in-module

        # Iterate through all known entry_points looking for self.scriptName
        matches = [ep for ep in metadata.entry_points()["console_scripts"] if ep.name == self.scriptName]
        if not matches:
            # TODO: This should an error once the integration tests modified to use pip install
            return self._func()  # pylint: disable=not-callable
            # raise NotImplementedError("Something is very wrong")

        # Call the entry_point from the extension with the highest priority
        rankedExtensions = extensionsByPriority()
        entrypoint = min(
            matches,
            key=lambda e: rankedExtensions.index(entrypointToExtension(e)),
        )
        entrypointFunc = entrypoint.load()

        # Check if entrypoint is DIRACScript
        if not isinstance(entrypointFunc, DIRACScript):
            raise ImportError(
                "Invalid dirac- console_scripts entry_point: "
                + repr(entrypoint)
                + "\n"
                + "All dirac- console_scripts should be wrapped in the DiracScript "
                + "decorator to ensure extension overlays are applied correctly."
            )
        return entrypointFunc._func()

    @classmethod
    def parseCommandLine(cls, script=False, ignoreErrors=False, initializeMonitor=False):
        """Parse command line

        :param str script: script name
        :param bool ignoreErrors: ignore errors when loading configuration
        :param bool initializeMonitor: to use monitoring
        """
        if not cls.alreadyInitialized:
            gLogger.showHeaders(False)
            cls.initialize(script, ignoreErrors, initializeMonitor, True)

        return (cls.localCfg.getUnprocessedSwitches(), cls.localCfg.getPositionalArguments())

    @classmethod
    def initialize(cls, script=False, ignoreErrors=False, initializeMonitor=False, enableCommandLine=False):
        """initialization

        :param str script: script name
        :param bool ignoreErrors: ignore errors when loading configuration
        :param bool initializeMonitor: to use monitoring
        :param bool enableCommandLine: enable parse command line
        """
        # Please do not call initialize in every file
        if cls.alreadyInitialized:
            return False
        userDisabled = not cls.localCfg.isCSEnabled()
        cls.alreadyInitialized = True
        if not userDisabled:
            cls.localCfg.disableCS()

        if not enableCommandLine:
            cls.localCfg.disableParsingCommandLine()

        if script:
            cls.scriptName = script
        cls.localCfg.setConfigurationForScript(cls.scriptName)

        if not ignoreErrors:
            cls.localCfg.addMandatoryEntry("/DIRAC/Setup")
        resultDict = cls.localCfg.loadUserData()
        if not ignoreErrors and not resultDict["OK"]:
            gLogger.error("There were errors when loading configuration", resultDict["Message"])
            sys.exit(1)
        if not userDisabled:
            cls.localCfg.enableCS()
        if initializeMonitor:
            gMonitor.setComponentType(gMonitor.COMPONENT_SCRIPT)
            gMonitor.setComponentName(cls.scriptName)
            gMonitor.setComponentLocation("script")
            gMonitor.initialize()
        else:
            gMonitor.disable()
        includeExtensionErrors()
        return True

    @classmethod
    def showHelp(cls, dummy=False, exitCode=0):
        """See :func:`~DIRAC.ConfigurationSystem.Client.LocalConfiguration.LocalConfiguration.showHelp`."""
        return cls.localCfg.showHelp(dummy=dummy, exitCode=exitCode)

    @classmethod
    def registerSwitches(cls, switches):
        """Register switches

        :param list switches: switches
        """
        for switch in switches:
            cls.registerSwitch(*switch)

    @classmethod
    def registerSwitch(cls, showKey, longKey, helpString, callback=False):
        """See :func:`~DIRAC.ConfigurationSystem.Client.LocalConfiguration.LocalConfiguration.registerCmdOpt`."""
        cls.localCfg.registerCmdOpt(showKey, longKey, helpString, callback)

    @classmethod
    def registerArgument(cls, description, mandatory=True, values=None, default=None):
        """See :func:`~DIRAC.ConfigurationSystem.Client.LocalConfiguration.LocalConfiguration.registerCmdArg`."""
        cls.localCfg.registerCmdArg(description, mandatory, values, default)

    @classmethod
    def getPositionalArgs(cls, group=False):
        """See :func:`~DIRAC.ConfigurationSystem.Client.LocalConfiguration.LocalConfiguration.getPositionalArguments`."""
        return cls.localCfg.getPositionalArguments(group)

    @classmethod
    def getExtraCLICFGFiles(cls):
        """See :func:`~DIRAC.ConfigurationSystem.Client.LocalConfiguration.LocalConfiguration.getExtraCLICFGFiles`."""
        return cls.localCfg.getExtraCLICFGFiles()

    @classmethod
    def getUnprocessedSwitches(cls):
        """See :func:`~DIRAC.ConfigurationSystem.Client.LocalConfiguration.LocalConfiguration.getUnprocessedSwitches`."""
        return cls.localCfg.getUnprocessedSwitches()

    @classmethod
    def addDefaultOptionValue(cls, option, value):
        """See :func:`~DIRAC.ConfigurationSystem.Client.LocalConfiguration.LocalConfiguration.addDefaultEntry`."""
        cls.localCfg.addDefaultEntry(option, value)

    @classmethod
    def setUsageMessage(cls, usageMessage):
        """See :func:`~DIRAC.ConfigurationSystem.Client.LocalConfiguration.LocalConfiguration.setUsageMessage`."""
        cls.localCfg.setUsageMessage(usageMessage)

    @classmethod
    def disableCS(cls):
        """See :func:`~DIRAC.ConfigurationSystem.Client.LocalConfiguration.LocalConfiguration.disableCS`."""
        cls.localCfg.disableCS()

    @classmethod
    def enableCS(cls):
        """See :func:`~DIRAC.ConfigurationSystem.Client.LocalConfiguration.LocalConfiguration.enableCS`."""
        return cls.localCfg.enableCS()



"""
It is used to test the database modules...
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import unittest

import DIRAC.MonitoringSystem.private.DBUtils as moduleTested

__RCSID__ = "$Id$"

################################################################################


class Test_DB(unittest.TestCase):
    """ """

    def setUp(self):
        """
        Setup
        """

        self.moduleTested = moduleTested
        self.testClass = self.moduleTested.DBUtils

    def tearDown(self):
        """
        Tear down
        """

        del self.moduleTested
        del self.testClass


################################################################################


class TestDbUtilsChain(Test_DB):
    """ """

    ################################################################################
    def test_instantiate(self):
        """tests that we can instantiate one object of the tested class"""

        module = self.testClass("MonitoringDB", "Test")
        self.assertEqual("DBUtils", module.__class__.__name__)

    ################################################################################
    def test_determineBucketSize(self):

        client = self.testClass("MonitoringDB", "Test")
        result = client._determineBucketSize(1458130176, 1458226213)
        self.assertTrue(result["OK"])
        self.assertEqual(result["Value"], ("30m", 1800))

        result = client._determineBucketSize(1458055269, 1458660069)
        self.assertTrue(result["OK"])
        self.assertEqual(result["Value"], ("2h", 7200))

        result = client._determineBucketSize(1458617047, 1458660247)
        self.assertTrue(result["OK"])
        self.assertEqual(result["Value"], ("10m", 600))

        result = client._determineBucketSize(1456068334, 1458660334)
        self.assertTrue(result["OK"])
        self.assertEqual(result["Value"], ("8h", 28800))

        result = client._determineBucketSize(1453476373, 1458660373)
        self.assertTrue(result["OK"])
        self.assertEqual(result["Value"], ("12h", 43200))

        result = client._determineBucketSize(1450884571, 1458660571)
        self.assertTrue(result["OK"])
        self.assertEqual(result["Value"], ("1d", 86400))

        result = client._determineBucketSize(1427038370, 1458660770)
        self.assertTrue(result["OK"])
        self.assertEqual(result["Value"], ("4d", 345600))

        result = client._determineBucketSize(1300807841, 1458660641)
        self.assertTrue(result["OK"])
        self.assertEqual(result["Value"], ("7d", 604800))


if __name__ == "__main__":
    testSuite = unittest.defaultTestLoader.loadTestsFromTestCase(Test_DB)
    testSuite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(TestDbUtilsChain))
    unittest.TextTestRunner(verbosity=2).run(testSuite)



#!/usr/bin/env python
########################################################################
# File :    dirac-wms-job-delete
# Author :  Stuart Paterson
########################################################################
"""
Peek StdOut of the given DIRAC job

Example:
  $ dirac-wms-job-peek 1
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(["JobID:    DIRAC Job ID"])
    _, args = Script.parseCommandLine(ignoreErrors=True)

    from DIRAC.Interfaces.API.Dirac import Dirac, parseArguments

    dirac = Dirac()
    exitCode = 0
    errorList = []

    for job in parseArguments(args):

        result = dirac.peekJob(job, printOutput=True)
        if not result["OK"]:
            errorList.append((job, result["Message"]))
            exitCode = 2

    for error in errorList:
        print("ERROR %s: %s" % error)

    DIRAC.exit(exitCode)


if __name__ == "__main__":
    main()



""" Test class for Job Agent
"""

# imports
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import pytest
from mock import MagicMock

# DIRAC Components
from DIRAC.WorkloadManagementSystem.Agent.JobAgent import JobAgent
from DIRAC import gLogger

gLogger.setLevel("DEBUG")

# Mock Objects
mockAM = MagicMock()
mockJM = MagicMock()
mockGCReply = MagicMock()
mockPMReply = MagicMock()
mockJW = MagicMock()
mockReply = MagicMock()


def test__getJDLParameters(mocker):
    """Testing JobAgent()._getJDLParameters()"""

    mocker.patch("DIRAC.WorkloadManagementSystem.Agent.JobAgent.AgentModule.__init__")

    jobAgent = JobAgent("Test", "Test1")
    jobAgent.log = gLogger
    jobAgent.log.setLevel("DEBUG")

    jdl = """
        [
            Executable = "dirac-jobexec";
            StdError = "std.err";
            LogLevel = "info";
            Site = "ANY";
            JobName = "helloWorld";
            Priority = "1";
            InputSandbox =
                {
                    "../../Integration/WorkloadManagementSystem/exe-script.py",
                    "exe-script.py",
                    "/tmp/tmpMQEink/jobDescription.xml",
                    "SB:FedericoSandboxSE|/SandBox/f/fstagni.lhcb_user/0c2/9f5/0c29f53a47d051742346b744c793d4d0.tar.bz2"
                };
            Arguments = "jobDescription.xml -o LogLevel=info";
            JobGroup = "lhcb";
            OutputSandbox =
                {
                    "helloWorld.log",
                    "std.err",
                    "std.out"
                };
            StdOutput = "std.out";
            InputData = "";
            JobType = "User";
            NumberOfProcessors = 16;
            Tags =
                {
                    "16Processors",
                    "MultiProcessor"
                };
        ]
        """

    result = jobAgent._getJDLParameters(jdl)

    assert result["OK"], result["Message"]
    assert result["Value"]["NumberOfProcessors"] == "16"
    assert result["Value"]["Tags"] == ["16Processors", "MultiProcessor"]


@pytest.mark.parametrize(
    "mockJMInput, expected",
    [
        ({"OK": True}, {"OK": True, "Value": "Problem Rescheduling Job"}),
        (
            {"OK": False, "Message": "Test"},
            {"OK": True, "Value": "Problem Rescheduling Job"},
        ),
    ],
)
def test__rescheduleFailedJob(mocker, mockJMInput, expected):
    """Testing JobAgent()._rescheduleFailedJob()"""

    mockJM.return_value = mockJMInput

    mocker.patch("DIRAC.WorkloadManagementSystem.Agent.JobAgent.AgentModule.__init__")

    jobAgent = JobAgent("Test", "Test1")

    jobID = 101
    message = "Test"

    jobAgent.log = gLogger
    jobAgent.log.setLevel("DEBUG")

    result = jobAgent._rescheduleFailedJob(jobID, message, stop=False)

    assert result == expected


@pytest.mark.parametrize(
    "mockGCReplyInput, mockPMReplyInput, expected",
    [
        (True, {"OK": True, "Value": "Test"}, {"OK": True, "Value": "Test"}),
        (
            True,
            {"OK": False, "Message": "Test"},
            {"OK": False, "Message": "Failed to setup proxy: Error retrieving proxy"},
        ),
        (
            False,
            {"OK": True, "Value": "Test"},
            {"OK": False, "Message": "Invalid Proxy"},
        ),
        (
            False,
            {"OK": False, "Message": "Test"},
            {"OK": False, "Message": "Invalid Proxy"},
        ),
    ],
)
def test__setupProxy(mocker, mockGCReplyInput, mockPMReplyInput, expected):
    """Testing JobAgent()._setupProxy()"""

    mockGCReply.return_value = mockGCReplyInput
    mockPMReply.return_value = mockPMReplyInput

    mocker.patch("DIRAC.WorkloadManagementSystem.Agent.JobAgent.AgentModule.__init__")
    mocker.patch("DIRAC.WorkloadManagementSystem.Agent.JobAgent.AgentModule", side_effect=mockAM)
    mocker.patch("DIRAC.WorkloadManagementSystem.Agent.JobAgent.gConfig.getValue", side_effect=mockGCReply)
    module_str = "DIRAC.WorkloadManagementSystem.Agent.JobAgent.gProxyManager.getPayloadProxyFromDIRACGroup"
    mocker.patch(module_str, side_effect=mockPMReply)

    jobAgent = JobAgent("Test", "Test1")

    ownerDN = "DIRAC"
    ownerGroup = "DIRAC"

    jobAgent.log = gLogger
    jobAgent.log.setLevel("DEBUG")

    result = jobAgent._setupProxy(ownerDN, ownerGroup)

    assert result["OK"] == expected["OK"]

    if result["OK"]:
        assert result["Value"] == expected["Value"]

    else:
        assert result["Message"] == expected["Message"]


def test__getCPUWorkLeft(mocker):
    """Testing JobAgent()._getCPUWorkLeft()"""

    mocker.patch("DIRAC.WorkloadManagementSystem.Agent.JobAgent.AgentModule.__init__")
    mocker.patch("DIRAC.WorkloadManagementSystem.Agent.JobAgent.AgentModule", side_effect=mockAM)

    jobAgent = JobAgent("Test", "Test1")
    jobAgent.log = gLogger
    jobAgent.log.setLevel("DEBUG")

    result = jobAgent._getCPUWorkLeft(0)

    assert 0 == result


@pytest.mark.parametrize(
    "mockGCReplyInput, mockPMReplyInput, expected",
    [
        (True, {"OK": True, "Value": "Test"}, {"OK": True, "Value": "Test"}),
        (
            True,
            {"OK": False, "Message": "Test"},
            {"OK": False, "Message": "Error retrieving proxy"},
        ),
        (False, {"OK": True, "Value": "Test"}, {"OK": True, "Value": "Test"}),
        (
            False,
            {"OK": False, "Message": "Test"},
            {"OK": False, "Message": "Error retrieving proxy"},
        ),
    ],
)
def test__requestProxyFromProxyManager(mocker, mockGCReplyInput, mockPMReplyInput, expected):
    """Testing JobAgent()._requestProxyFromProxyManager()"""

    mockGCReply.return_value = mockGCReplyInput
    mockPMReply.return_value = mockPMReplyInput

    mocker.patch("DIRAC.WorkloadManagementSystem.Agent.JobAgent.AgentModule.__init__")
    mocker.patch("DIRAC.WorkloadManagementSystem.Agent.JobAgent.AgentModule", side_effect=mockAM)
    mocker.patch("DIRAC.WorkloadManagementSystem.Agent.JobAgent.gConfig.getValue", side_effect=mockGCReply)
    module_str = "DIRAC.WorkloadManagementSystem.Agent.JobAgent.gProxyManager.getPayloadProxyFromDIRACGroup"
    mocker.patch(module_str, side_effect=mockPMReply)

    jobAgent = JobAgent("Test", "Test1")

    ownerDN = "DIRAC"
    ownerGroup = "DIRAC"

    jobAgent.log = gLogger
    jobAgent.log.setLevel("DEBUG")

    result = jobAgent._requestProxyFromProxyManager(ownerDN, ownerGroup)

    assert result["OK"] == expected["OK"]

    if result["OK"]:
        assert result["Value"] == expected["Value"]

    else:
        assert result["Message"] == expected["Message"]


def test__checkInstallSoftware(mocker):
    """Testing JobAgent()._checkInstallSoftware()"""

    mocker.patch("DIRAC.WorkloadManagementSystem.Agent.JobAgent.AgentModule.__init__")

    jobAgent = JobAgent("Test", "Test1")
    jobAgent.log = gLogger
    jobAgent.log.setLevel("DEBUG")

    result = jobAgent._checkInstallSoftware(101, {}, {}, MagicMock())

    assert result["OK"], result["Message"]
    assert result["Value"] == "Job has no software installation requirement"


@pytest.mark.parametrize(
    "mockJWInput, expected", [({"OK": False, "Message": "Test"}, {"OK": False, "Message": "Test"})]
)
def test_submitJob(mocker, mockJWInput, expected):
    """Testing JobAgent()._submitJob()"""

    mockJW.return_value = mockJWInput

    mocker.patch("DIRAC.WorkloadManagementSystem.Agent.JobAgent.AgentModule.__init__")
    mocker.patch(
        "DIRAC.WorkloadManagementSystem.Agent.JobAgent.AgentModule._AgentModule__moduleProperties",
        side_effect=lambda x, y=None: y,
        create=True,
    )
    mocker.patch("DIRAC.WorkloadManagementSystem.Agent.JobAgent.createJobWrapper", side_effect=mockJW)

    jobAgent = JobAgent("Test", "Test1")
    jobAgent.log = gLogger
    jobAgent.log.setLevel("DEBUG")
    jobAgent.ceName = "Test"

    result = jobAgent._submitJob(101, {}, {}, {}, "", 1)

    assert result["OK"] == expected["OK"]

    if not result["OK"]:
        assert result["Message"] == expected["Message"]



"""  Proxy Renewal agent is the key element of the Proxy Repository
     which maintains the user proxies alive

    .. literalinclude:: ../ConfigTemplate.cfg
      :start-after: ##BEGIN MyProxyRenewalAgent
      :end-before: ##END
      :dedent: 2
      :caption: MyProxyRenewalAgent options
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import concurrent.futures

from DIRAC import gLogger, S_OK

from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.FrameworkSystem.DB.ProxyDB import ProxyDB

DEFAULT_MAIL_FROM = "proxymanager@diracgrid.org"


class MyProxyRenewalAgent(AgentModule):
    def initialize(self):

        requiredLifeTime = self.am_getOption("MinimumLifeTime", 3600)
        renewedLifeTime = self.am_getOption("RenewedLifeTime", 54000)
        mailFrom = self.am_getOption("MailFrom", DEFAULT_MAIL_FROM)
        self.proxyDB = ProxyDB(useMyProxy=True, mailFrom=mailFrom)

        gLogger.info("Minimum Life time      : %s" % requiredLifeTime)
        gLogger.info("Life time on renew     : %s" % renewedLifeTime)
        gLogger.info("MyProxy server         : %s" % self.proxyDB.getMyProxyServer())
        gLogger.info("MyProxy max proxy time : %s" % self.proxyDB.getMyProxyMaxLifeTime())

        return S_OK()

    def __renewProxyForCredentials(self, userDN, userGroup):
        lifeTime = self.am_getOption("RenewedLifeTime", 54000)
        gLogger.info("Renewing for %s@%s %s secs" % (userDN, userGroup, lifeTime))
        retVal = self.proxyDB.renewFromMyProxy(userDN, userGroup, lifeTime=lifeTime)
        if not retVal["OK"]:
            gLogger.error("Failed to renew proxy", "for %s@%s : %s" % (userDN, userGroup, retVal["Message"]))
        else:
            gLogger.info("Renewed proxy for %s@%s" % (userDN, userGroup))

    def __treatRenewalCallback(self, oTJ, exceptionList):
        gLogger.exception(lException=exceptionList)

    def execute(self):
        """The main agent execution method"""
        self.proxyDB.purgeLogs()
        gLogger.info("Purging expired requests")
        retVal = self.proxyDB.purgeExpiredRequests()
        if retVal["OK"]:
            gLogger.info(" purged %s requests" % retVal["Value"])
        gLogger.info("Purging expired proxies")
        retVal = self.proxyDB.purgeExpiredProxies()
        if retVal["OK"]:
            gLogger.info(" purged %s proxies" % retVal["Value"])
        retVal = self.proxyDB.getCredentialsAboutToExpire(self.am_getOption("MinimumLifeTime", 3600))
        if not retVal["OK"]:
            return retVal
        data = retVal["Value"]
        gLogger.info("Renewing %s proxies..." % len(data))
        with concurrent.futures.ThreadPoolExecutor(max_workers=10) as executor:
            futures = []
            for record in data:
                userDN = record[0]
                userGroup = record[1]
                futures.append(executor.submit(self.__renewProxyForCredentials, userDN, userGroup))
        return S_OK()



""" AlwaysProbingPolicy module """
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC import S_OK
from DIRAC.ResourceStatusSystem.PolicySystem.PolicyBase import PolicyBase

__RCSID__ = "$Id$"


class AlwaysProbingPolicy(PolicyBase):
    """
    The AlwaysProbingPolicy is a dummy module that can be used as example, it
    always returns Probing status.
    """

    @staticmethod
    def _evaluate(commandResult):
        """
        It returns Probing status, evaluates the default command, but its output
        is completely ignored.
        """

        policyResult = {"Status": "Probing", "Reason": "AlwaysProbing"}

        return S_OK(policyResult)



#!/usr/bin/env python
"""
This is a simple script that can be used for synchronizing pilot files
to the current directory.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os
import json
import hashlib


from DIRAC import S_OK
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


includeMasterCS = True


def setNoMasterCS(optVal):
    global includeMasterCS
    includeMasterCS = False
    return S_OK()


@Script()
def main():
    global includeMasterCS
    Script.registerSwitch("n", "noMasterCS", "do not include master CS", setNoMasterCS)
    Script.parseCommandLine()

    from DIRAC import gLogger, exit as DIRACExit
    from DIRAC.WorkloadManagementSystem.Utilities.PilotCStoJSONSynchronizer import PilotCStoJSONSynchronizer

    ps = PilotCStoJSONSynchronizer()

    gLogger.verbose("Parameters for this sync:")
    gLogger.verbose("repo=" + ps.pilotRepo)
    gLogger.verbose("VO repo=" + ps.pilotVORepo)
    gLogger.verbose("projectDir=" + ps.projectDir)
    gLogger.verbose("pilotScriptsPath=" + ps.pilotScriptPath)
    gLogger.verbose("pilotVOScriptsPath=" + ps.pilotVOScriptPath)
    gLogger.verbose("pilotRepoBranch=" + ps.pilotRepoBranch)
    gLogger.verbose("pilotVORepoBranch=" + ps.pilotVORepoBranch)

    # pilot.json
    res = ps.getCSDict(includeMasterCS=includeMasterCS)
    if not res["OK"]:
        DIRACExit(1)
    pilotDict = res["Value"]
    print(json.dumps(pilotDict, indent=4, sort_keys=True))  # just print here as formatting is important
    with open("pilot.json", "w") as jf:
        json.dump(pilotDict, jf)

    # pilot files
    res = ps.syncScripts()
    if not res["OK"]:
        DIRACExit(1)
    gLogger.always(res["Value"])
    tarPath, tarFiles = res["Value"]

    allFiles = [tarPath] + tarFiles + ["pilot.json"]

    # checksums
    checksumDict = {}
    for pFile in allFiles:
        filename = os.path.basename(pFile)
        with open(pFile, "rb") as fp:
            checksumDict[filename] = hashlib.sha512(fp.read()).hexdigest()
        cksPath = "checksums.sha512"
    with open(cksPath, "wt") as chksums:
        for filename, chksum in sorted(checksumDict.items()):
            # same as the output from sha512sum commands
            chksums.write("%s  %s\n" % (chksum, filename))

    allFiles = allFiles + [cksPath]

    print(allFiles)


if __name__ == "__main__":
    main()



""" The gateway service is used for forwarding service calls to the appropriate services.

    For this to be used, the following CS option is required::

      DIRAC
      {
        Gateways
        {
          my.site.org = dips://thisIsAn.url.org:9159/Framework/Gateway
        }
      }

    At the same time, this same gateway service should be run with option /LocalInstallation/Site
    which is different from "my.site.org" or whatever is set in the option above, to avoid initialization loops.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


import sys
from six import BytesIO

from concurrent.futures import ThreadPoolExecutor

import DIRAC
from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Core.Utilities.LockRing import LockRing
from DIRAC.Core.Utilities.DictCache import DictCache
from DIRAC.ConfigurationSystem.Client.ConfigurationData import gConfigurationData
from DIRAC.Core.DISET.private.FileHelper import FileHelper
from DIRAC.Core.DISET.private.MessageBroker import MessageBroker, getGlobalMessageBroker
from DIRAC.Core.DISET.MessageClient import MessageClient
from DIRAC.Core.Security.X509Chain import X509Chain  # pylint: disable=import-error
from DIRAC.Core.DISET.private.Service import Service
from DIRAC.Core.DISET.RPCClient import RPCClient
from DIRAC.Core.DISET.TransferClient import TransferClient
from DIRAC.Core.DISET.private.BaseClient import BaseClient

__RCSID__ = "$Id$"


class GatewayService(Service):
    """Inherits from Service so it can (and should) be run as a DIRAC service,
    but replaces several of the internal methods
    """

    GATEWAY_NAME = "Framework/Gateway"

    def __init__(self):
        """Initialize like a real service"""
        super(GatewayService, self).__init__(
            {
                "modName": GatewayService.GATEWAY_NAME,
                "loadName": GatewayService.GATEWAY_NAME,
                "standalone": True,
                "moduleObj": sys.modules[DIRAC.Core.DISET.private.GatewayService.GatewayService.__module__],
                "classObj": self.__class__,
            }
        )
        self.__delegatedCredentials = DictCache()
        self.__transferBytesLimit = 1024 * 1024 * 100
        # to be resolved
        self._url = None
        self._handler = None
        self._threadPool = None
        self._msgBroker = None
        self._msgForwarder = None

    def initialize(self):
        """This replaces the standard initialize from Service"""
        # Build the URLs
        self._url = self._cfg.getURL()
        if not self._url:
            return S_ERROR("Could not build service URL for %s" % GatewayService.GATEWAY_NAME)
        gLogger.verbose("Service URL is %s" % self._url)
        # Load handler
        result = self._loadHandlerInit()
        if not result["OK"]:
            return result
        self._handler = result["Value"]
        # Discover Handler
        self._threadPool = ThreadPoolExecutor(max(0, self._cfg.getMaxThreads()))

        self._msgBroker = MessageBroker("%sMSB" % GatewayService.GATEWAY_NAME, threadPool=self._threadPool)
        self._msgBroker.useMessageObjects(False)
        getGlobalMessageBroker().useMessageObjects(False)
        self._msgForwarder = MessageForwarder(self._msgBroker)
        return S_OK()

    def _processInThread(self, clientTransport):
        """Threaded process function"""
        # Handshake
        try:
            clientTransport.handshake()
        except Exception:
            return
        # Add to the transport pool
        trid = self._transportPool.add(clientTransport)
        if not trid:
            return
        # Receive and check proposal
        result = self._receiveAndCheckProposal(trid)
        if not result["OK"]:
            self._transportPool.sendAndClose(trid, result)
            return
        proposalTuple = result["Value"]
        # Instantiate handler
        result = self.__getClientInitArgs(trid, proposalTuple)
        if not result["OK"]:
            self._transportPool.sendAndClose(trid, result)
            return
        clientInitArgs = result["Value"]
        # Execute the action
        result = self._processProposal(trid, proposalTuple, clientInitArgs)
        # Close the connection if required
        if result["closeTransport"]:
            self._transportPool.close(trid)
        return result

    def _receiveAndCheckProposal(self, trid):
        clientTransport = self._transportPool.get(trid)
        # Get the peer credentials
        credDict = clientTransport.getConnectingCredentials()
        # Receive the action proposal
        retVal = clientTransport.receiveData(1024)
        if not retVal["OK"]:
            gLogger.error(
                "Invalid action proposal",
                "%s %s" % (self._createIdentityString(credDict, clientTransport), retVal["Message"]),
            )
            return S_ERROR("Invalid action proposal")
        proposalTuple = retVal["Value"]
        gLogger.debug("Received action from client", "/".join(list(proposalTuple[1])))
        # Check if there are extra credentials
        if proposalTuple[2]:
            clientTransport.setExtraCredentials(proposalTuple[2])
        return S_OK(proposalTuple)

    def __getClientInitArgs(self, trid, proposalTuple):
        clientTransport = self._transportPool.get(trid)
        # Get the peer credentials
        credDict = clientTransport.getConnectingCredentials()
        if "x509Chain" not in credDict:
            return S_OK()
        cKey = (
            credDict["DN"],
            credDict.get("group", False),
            credDict.get("extraCredentials", False),
            credDict["isLimitedProxy"],
        )
        dP = self.__delegatedCredentials.get(cKey, 3600)
        idString = self._createIdentityString(credDict, clientTransport)
        if dP:
            gLogger.verbose("Proxy for %s is cached" % idString)
            return S_OK(dP)
        result = self.__requestDelegation(clientTransport, credDict)
        if not result["OK"]:
            gLogger.warn("Could not get proxy for %s: %s" % (idString, result["Message"]))
            return result
        delChain = result["Value"]
        delegatedChain = delChain.dumpAllToString()["Value"]
        secsLeft = delChain.getRemainingSecs()["Value"] - 1
        clientInitArgs = {
            BaseClient.KW_SETUP: proposalTuple[0][1],
            BaseClient.KW_TIMEOUT: 600,
            BaseClient.KW_IGNORE_GATEWAYS: True,
            BaseClient.KW_USE_CERTIFICATES: False,
            BaseClient.KW_PROXY_STRING: delegatedChain,
        }
        if BaseClient.KW_EXTRA_CREDENTIALS in credDict:
            clientInitArgs[BaseClient.KW_EXTRA_CREDENTIALS] = credDict[BaseClient.KW_EXTRA_CREDENTIALS]
        gLogger.warn("Got delegated proxy for %s: %s secs left" % (idString, secsLeft))
        self.__delegatedCredentials.add(cKey, secsLeft, clientInitArgs)
        return S_OK(clientInitArgs)

    def __requestDelegation(self, clientTransport, credDict):
        peerChain = credDict["x509Chain"]
        retVal = peerChain.getCertInChain()["Value"].generateProxyRequest()
        if not retVal["OK"]:
            return retVal
        delegationRequest = retVal["Value"]
        retVal = delegationRequest.dumpRequest()
        if not retVal["OK"]:
            retVal = S_ERROR("Server Error: Can't generate delegation request")
            clientTransport.sendData(retVal)
            return retVal
        gLogger.info("Sending delegation request for %s" % delegationRequest.getSubjectDN()["Value"])
        clientTransport.sendData(S_OK({"delegate": retVal["Value"]}))
        delegatedCertChain = clientTransport.receiveData()
        delegatedChain = X509Chain(keyObj=delegationRequest.getPKey())
        retVal = delegatedChain.loadChainFromString(delegatedCertChain)
        if not retVal["OK"]:
            retVal = S_ERROR("Error in receiving delegated proxy: %s" % retVal["Message"])
            clientTransport.sendData(retVal)
            return retVal
        return S_OK(delegatedChain)

    # Msg

    def _mbConnect(self, trid, handlerObj=None):
        return S_OK()

    def _mbReceivedMsg(self, cliTrid, msgObj):
        return self._msgForwarder.msgFromClient(cliTrid, msgObj)

    def _mbDisconnect(self, cliTrid):
        self._msgForwarder.cliDisconnect(cliTrid)

    # Execute action

    def _executeAction(self, trid, proposalTuple, clientInitArgs):
        clientTransport = self._transportPool.get(trid)
        credDict = clientTransport.getConnectingCredentials()
        targetService = proposalTuple[0][0]
        actionType = proposalTuple[1][0]
        actionMethod = proposalTuple[1][1]
        idString = self._createIdentityString(credDict, clientTransport)
        # OOkay! Lets do the magic!
        retVal = clientTransport.receiveData()
        if not retVal["OK"]:
            gLogger.error("Error while receiving file description", retVal["Message"])
            clientTransport.sendData(S_ERROR("Error while receiving file description: %s" % retVal["Message"]))
            return
        if actionType == "FileTransfer":
            gLogger.warn("Received a file transfer action from %s" % idString)
            clientTransport.sendData(S_OK("Accepted"))
            retVal = self.__forwardFileTransferCall(
                targetService, clientInitArgs, actionMethod, retVal["Value"], clientTransport
            )
        elif actionType == "RPC":
            gLogger.info("Forwarding %s/%s action to %s for %s" % (actionType, actionMethod, targetService, idString))
            retVal = self.__forwardRPCCall(targetService, clientInitArgs, actionMethod, retVal["Value"])
        elif actionType == "Connection" and actionMethod == "new":
            gLogger.info("Initiating a messaging connection to %s for %s" % (targetService, idString))
            retVal = self._msgForwarder.addClient(trid, targetService, clientInitArgs, retVal["Value"])
        else:
            gLogger.warn("Received an invalid %s/%s action from %s" % (actionType, actionMethod, idString))
            retVal = S_ERROR("Unknown type of action (%s)" % actionType)
        # TODO: Send back the data?
        if "rpcStub" in retVal:
            retVal.pop("rpcStub")
        clientTransport.sendData(retVal)
        return retVal

    def __forwardRPCCall(self, targetService, clientInitArgs, method, params):
        if targetService == "Configuration/Server":
            if method == "getCompressedDataIfNewer":
                # Relay CS data directly
                serviceVersion = gConfigurationData.getVersion()
                retDict = {"newestVersion": serviceVersion}
                clientVersion = params[0]
                if clientVersion < serviceVersion:
                    retDict["data"] = gConfigurationData.getCompressedData()
                return S_OK(retDict)
        # Default
        rpcClient = RPCClient(targetService, **clientInitArgs)
        methodObj = getattr(rpcClient, method)
        return methodObj(*params)

    def __forwardFileTransferCall(self, targetService, clientInitArgs, method, params, clientTransport):
        transferRelay = TransferRelay(targetService, **clientInitArgs)
        transferRelay.setTransferLimit(self.__transferBytesLimit)
        cliFH = FileHelper(clientTransport)
        # Check file size
        if method.find("ToClient") > -1:
            cliFH.setDirection("send")
        elif method.find("FromClient") > -1:
            cliFH.setDirection("receive")
            if not self.__ftCheckMaxTransferSize(params[2]):
                cliFH.markAsTransferred()
                return S_ERROR("Transfer size is too big")
        # Forward queries
        try:
            relayMethodObject = getattr(transferRelay, "forward%s" % method)
        except Exception:
            return S_ERROR("Cannot forward unknown method %s" % method)
        result = relayMethodObject(cliFH, params)
        return result

    def __ftCheckMaxTransferSize(self, requestedTransferSize):
        if not self.__transferBytesLimit:
            return True
        if not requestedTransferSize:
            return True
        if requestedTransferSize <= self.__transferBytesLimit:
            return True
        return False


class TransferRelay(TransferClient):
    def setTransferLimit(self, trLimit):
        self.__transferBytesLimit = trLimit
        self.__currentMethod = ""

    def infoMsg(self, msg, dynMsg=""):
        gLogger.info("[%s] %s" % (self.__currentMethod, msg), dynMsg)

    def errMsg(self, msg, dynMsg=""):
        gLogger.error("[%s] %s" % (self.__currentMethod, msg), dynMsg)

    def getDataFromClient(self, clientFileHelper):
        sIO = BytesIO()
        self.infoMsg("About to get data from client")
        result = clientFileHelper.networkToDataSink(sIO, self.__transferBytesLimit)
        if not result["OK"]:
            sIO.close()
            self.errMsg("Could not get data from client", result["Message"])
            return result
        data = sIO.getvalue()
        sIO.close()
        self.infoMsg("Got %s bytes from client" % len(data))
        return S_OK(data)

    def sendDataToClient(self, clientFileHelper, dataToSend):
        self.infoMsg("About to get send data to client")
        result = clientFileHelper.BufferToNetwork(dataToSend)
        if not result["OK"]:
            self.errMsg("Could not send data to client", result["Message"])
            return result
        self.infoMsg("Sent %s bytes from client" % len(dataToSend))
        return S_OK()

    def sendDataToService(self, srvMethod, params, data):
        self.infoMsg("Sending header request to %s" % self.getDestinationService(), str(params))
        result = self._sendTransferHeader(srvMethod, params)
        if not result["OK"]:
            self.errMsg("Could not send header", result["Message"])
            return result
        self.infoMsg("Starting to send data to service")
        _, srvTransport = result["Value"]
        srvFileHelper = FileHelper(srvTransport)
        srvFileHelper.setDirection("send")
        result = srvFileHelper.BufferToNetwork(data)
        if not result["OK"]:
            self.errMsg("Could send data to server", result["Message"])
            srvTransport.close()
            return result
        self.infoMsg("Data sent to service (%s bytes)" % len(data))
        retVal = srvTransport.receiveData()
        srvTransport.close()
        return retVal

    def getDataFromService(self, srvMethod, params):
        self.infoMsg("Sending header request to %s" % self.getDestinationService(), str(params))
        result = self._sendTransferHeader(srvMethod, params)
        if not result["OK"]:
            self.errMsg("Could not send header", result["Message"])
            return result
        self.infoMsg("Starting to receive data from service")
        _, srvTransport = result["Value"]
        srvFileHelper = FileHelper(srvTransport)
        srvFileHelper.setDirection("receive")
        sIO = BytesIO()
        result = srvFileHelper.networkToDataSink(sIO, self.__transferBytesLimit)
        if not result["OK"]:
            self.errMsg("Could not receive data from server", result["Message"])
            srvTransport.close()
            sIO.close()
            return result
        dataReceived = sIO.getvalue()
        sIO.close()
        self.infoMsg("Received %s bytes from service" % len(dataReceived))
        retVal = srvTransport.receiveData()
        srvTransport.close()
        if not retVal["OK"]:
            return retVal
        return S_OK({"data": dataReceived, "srvResponse": retVal})

    def forwardFromClient(self, clientFileHelper, params):
        fileId, token = params[:2]
        self.__currentMethod = "FromClient"
        result = self.getDataFromClient(clientFileHelper)
        if not result["OK"]:
            return result
        dataReceived = result["Value"]
        receivedBytes = clientFileHelper.getTransferedBytes()
        return self.sendDataToService("FromClient", (fileId, token, receivedBytes), dataReceived)

    def forwardBulkFromClient(self, clientFileHelper, params):
        fileId, token = params[:2]
        self.__currentMethod = "BulkFromClient"
        result = self.getDataFromClient(clientFileHelper)
        if not result["OK"]:
            return result
        dataReceived = result["Value"]
        receivedBytes = clientFileHelper.getTransferedBytes()
        return self.sendDataToService("BulkFromClient", (fileId, token, receivedBytes), dataReceived)

    def forwardToClient(self, clientFileHelper, params):
        fileId, token = params[:2]
        self.__currentMethod = "ToClient"
        result = self.getDataFromService("ToClient", (fileId, token))
        if not result["OK"]:
            return result
        dataReceived = result["Value"]["data"]
        srvResponse = result["Value"]["srvResponse"]
        result = self.sendDataToClient(clientFileHelper, dataReceived)
        if not result["OK"]:
            return result
        return srvResponse

    def forwardBulkToClient(self, clientFileHelper, params):
        fileId, token = params[:2]
        self.__currentMethod = "BulkToClient"
        result = self.getDataFromService("BulkToClient", (fileId, token))
        if not result["OK"]:
            return result
        dataReceived = result["Value"]["data"]
        srvResponse = result["Value"]["srvResponse"]
        result = self.sendDataToClient(clientFileHelper, dataReceived)
        if not result["OK"]:
            return result
        return srvResponse

    def forwardListBulk(self, clientFileHelper, params):
        self.__currentMethod = "ListBulk"
        self.infoMsg("Sending header request to %s" % self.getDestinationService(), str(params))
        result = self._sendTransferHeader("ListBulk", params)
        if not result["OK"]:
            self.errMsg("Could not send header", result["Message"])
            return result
        _, srvTransport = result["Value"]
        response = srvTransport.receiveData(1048576)
        srvTransport.close()
        self.infoMsg("Sending data back to client")
        return response


class MessageForwarder(object):
    def __init__(self, msgBroker):
        self.__inOutLock = LockRing().getLock()
        self.__msgBroker = msgBroker
        self.__byClient = {}
        self.__srvToCliTrid = {}

    def addClient(self, cliTrid, destination, clientInitParams, connectParams):
        if cliTrid in self.__byClient:
            gLogger.fatal("Trid is duplicated!! this shouldn't happen")
            return
        msgClient = MessageClient(destination, **clientInitParams)
        msgClient.subscribeToDisconnect(self.__srvDisconnect)
        msgClient.subscribeToAllMessages(self.msgFromSrv)
        msgClient.setUniqueName(connectParams[0])
        result = msgClient.connect(**connectParams[1])
        if not result["OK"]:
            return result
        self.__inOutLock.acquire()
        try:
            self.__byClient[cliTrid] = {"srvEnd": msgClient, "srvTrid": msgClient.getTrid(), "srvName": destination}
            self.__srvToCliTrid[msgClient.getTrid()] = cliTrid
        finally:
            self.__inOutLock.release()
        return result

    def __srvDisconnect(self, srvEndCli):
        try:
            cliTrid = self.__srvToCliTrid[srvEndCli.getTrid()]
        except IndexError:
            gLogger.exception("This shouldn't happen!")
        gLogger.info("Service %s disconnected messaging connection" % self.__byClient[cliTrid]["srvName"])
        self.__msgBroker.removeTransport(cliTrid)
        self.__removeClient(cliTrid)

    def cliDisconnect(self, cliTrid):
        if cliTrid not in self.__byClient:
            gLogger.fatal("This shouldn't happen!")
            return
        gLogger.info("Client to %s disconnected messaging connection" % self.__byClient[cliTrid]["srvName"])
        self.__byClient[cliTrid]["srvEnd"].disconnect()
        self.__removeClient(cliTrid)

    def __removeClient(self, cliTrid):
        self.__inOutLock.acquire()
        try:
            try:
                srvTrid = self.__byClient[cliTrid]["srvTrid"]
                self.__byClient.pop(cliTrid)
                self.__srvToCliTrid.pop(srvTrid)
            except Exception as e:
                gLogger.exception("This shouldn't happen!", e)
        finally:
            self.__inOutLock.release()

    def msgFromClient(self, cliTrid, msgObj):
        gLogger.info("Message %s to %s service" % (msgObj.getName(), self.__byClient[cliTrid]["srvName"]))
        result = self.__byClient[cliTrid]["srvEnd"].sendMessage(msgObj)
        return result

    def msgFromSrv(self, srvEndCli, msgObj):
        try:
            cliTrid = self.__srvToCliTrid[srvEndCli.getTrid()]
        except Exception:
            gLogger.exception("This shouldn't happen")
            return S_ERROR("MsgFromSrv -> Mismatched srv2cli trid")
        gLogger.info("Message %s from %s service" % (msgObj.getName(), self.__byClient[cliTrid]["srvName"]))
        return self.__msgBroker.sendMessage(cliTrid, msgObj)



"""TransformationPlugin is a class wrapping the supported transformation plugins

New plugins can be created by defining a function.
The function name has to be ``_MyPlugin``, and the plugin is then call ``'MyPlugin'``. Do not forget to enable new
plugins in the ``Operations/Transformation/AllowedPlugins`` list.
The return value of the function has to be ``S_OK`` with a list of tuples. Each Tuple is a pair of a StorageElement
and list of LFNs to treat in given task::

  return S_OK([('SE_1', [lfn_1_1, lfn_1_2, ...]),
               ('SE_1', [lfn_2_1, lfn_2_2, ...]),
               # ...
               ('SE_I', [lfn_J_1, lfn_J_2, ...]),
               ])

Inside the plugin function, the relevant LFNs can be accessed in the ``self.data`` dictionary, and transformation
parameters are obtained contained in the ``self.params`` dictionary. See also the
:class:`~DIRAC.TransformationSystem.Client.Utilities.PluginUtilities` class.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import random
import time

from DIRAC import gConfig, S_OK, S_ERROR
from DIRAC.Core.Utilities.SiteSEMapping import getSitesForSE, getSEsForSite
from DIRAC.Core.Utilities.List import breakListIntoChunks

from DIRAC.Resources.Catalog.FileCatalog import FileCatalog
from DIRAC.DataManagementSystem.Client.DataManager import DataManager
from DIRAC.TransformationSystem.Client.PluginBase import PluginBase
from DIRAC.TransformationSystem.Client.Utilities import PluginUtilities, getFileGroups
from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient

__RCSID__ = "$Id$"


class TransformationPlugin(PluginBase):
    """A TransformationPlugin object should be instantiated by every transformation.

    :param str plugin: A plugin name has to be passed in: it will then be executed as one of the functions below, e.g.
       plugin = 'BySize' will execute TransformationPlugin('BySize')._BySize()

    :param transClient: TransformationManagerClient instance
    :param dataManager: DataManager instance
    :param fc: FileCatalog instance
    """

    def __init__(self, plugin, transClient=None, dataManager=None, fc=None):
        """Constructor of the TransformationPlugin.

        Instantiate clients, if not given, and set up the PluginUtilities.
        """
        super(TransformationPlugin, self).__init__(plugin)

        self.data = {}
        self.files = False
        self.startTime = time.time()
        self.valid = False

        if transClient is None:
            transClient = TransformationClient()

        if dataManager is None:
            dataManager = DataManager()

        if fc is None:
            fc = FileCatalog()

        self.util = PluginUtilities(plugin, transClient=transClient, dataManager=dataManager, fc=fc)

    def __del__(self):
        """Destructor: print the elapsed time"""
        self.util.logInfo("Execution finished, timing: %.3f seconds" % (time.time() - self.startTime))

    def isOK(self):
        """Check if all information is present"""
        self.valid = True
        if (not self.data) or (not self.params):
            self.valid = False
        return self.valid

    def setParameters(self, params):
        """Need to pass parameters also to self.util"""
        self.params = params
        self.util.setParameters(params)

    def setInputData(self, data):
        """Set the replica information as data member"""
        self.data = data

    def setTransformationFiles(self, files):
        """Set the TS files as data member"""
        self.files = files

    def _Standard(self):
        """Simply group by replica location (if any)"""
        return self.util.groupByReplicas(self.data, self.params["Status"])

    def _BySize(self):
        """Alias for groupBySize"""
        return self._groupBySize()

    def _groupBySize(self, files=None):
        """
        Generate a task for a given amount of data at a (set of) SE
        """
        if not files:
            files = self.data
        else:
            files = dict(zip(files, [self.data[lfn] for lfn in files]))
        return self.util.groupBySize(files, self.params["Status"])

    def _Broadcast(self):
        """This plug-in takes files found at the sourceSE and broadcasts to all (or a selection of) targetSEs.

        Parameters used by this plugin:

        * SourceSE: Optional: only files at this location are treated
        * TargetSE: Where to broadcast files to
        * Destinations: Optional: integer, files are only broadcast to this number of TargetSEs, Destinations has to be
          larger than the number of TargetSEs
        * GroupSize: number of files per task
        """
        if not self.params:
            return S_ERROR("TransformationPlugin._Broadcast: The 'Broadcast' plugin requires additional parameters.")

        sourceSEs = set(self.util.seParamtoList(self.params.get("SourceSE", [])))
        targetSEs = self.util.seParamtoList(self.params["TargetSE"])
        destinations = int(self.params.get("Destinations", 0))
        if destinations and (destinations >= len(targetSEs)):
            destinations = 0

        status = self.params["Status"]
        groupSize = self.params["GroupSize"]  # Number of files per tasks

        fileGroups = getFileGroups(self.data)  # groups by SE
        targetSELfns = {}
        for replicaSE, lfns in fileGroups.items():
            ses = replicaSE.split(",")
            atSource = (not sourceSEs) or set(ses).intersection(sourceSEs)
            if not atSource:
                continue

            for lfn in lfns:
                targets = []
                sourceSites = self._getSitesForSEs(ses)
                random.shuffle(targetSEs)
                for targetSE in targetSEs:
                    site = self._getSiteForSE(targetSE)["Value"]
                    if site not in sourceSites:
                        if (destinations) and (len(targets) >= destinations):
                            continue
                        sourceSites.append(site)
                    targets.append(targetSE)  # after all, if someone wants to copy to the source, it's his choice
                strTargetSEs = ",".join(sorted(targets))
                targetSELfns.setdefault(strTargetSEs, []).append(lfn)
        tasks = []
        for ses, lfns in targetSELfns.items():
            tasksLfns = breakListIntoChunks(lfns, groupSize)
            for taskLfns in tasksLfns:
                if (status == "Flush") or (len(taskLfns) >= int(groupSize)):
                    # do not allow groups smaller than the groupSize, except if transformation is in flush state
                    tasks.append((ses, taskLfns))
        return S_OK(tasks)

    def _ByShare(self, shareType="CPU"):
        """first get the shares from the CS, and then makes the grouping looking at the history"""
        res = self._getShares(shareType, normalise=True)
        if not res["OK"]:
            return res
        cpuShares = res["Value"]
        self.util.logInfo("Obtained the following target shares (%):")
        for site in sorted(cpuShares):
            self.util.logInfo("%s: %.1f" % (site.ljust(15), cpuShares[site]))

        # Get the existing destinations from the transformationDB
        res = self.util.getExistingCounters(requestedSites=list(cpuShares))
        if not res["OK"]:
            self.util.logError("Failed to get existing file share", res["Message"])
            return res
        existingCount = res["Value"]
        if existingCount:
            self.util.logInfo("Existing site utilization (%):")
            normalisedExistingCount = self.util._normaliseShares(
                existingCount.copy()
            )  # pylint: disable=protected-access
            for se in sorted(normalisedExistingCount):
                self.util.logInfo("%s: %.1f" % (se.ljust(15), normalisedExistingCount[se]))

        # Group the input files by their existing replicas
        res = self.util.groupByReplicas(self.data, self.params["Status"])
        if not res["OK"]:
            return res
        replicaGroups = res["Value"]

        tasks = []
        # For the replica groups
        for replicaSE, lfns in replicaGroups:
            possibleSEs = replicaSE.split(",")
            # Determine the next site based on requested shares, existing usage and candidate sites
            res = self._getNextSite(existingCount, cpuShares, candidates=self._getSitesForSEs(possibleSEs))
            if not res["OK"]:
                self.util.logError("Failed to get next destination SE", res["Message"])
                continue
            targetSite = res["Value"]
            # Resolve the ses for the target site
            res = getSEsForSite(targetSite)
            if not res["OK"]:
                continue
            ses = res["Value"]
            # Determine the selected SE and create the task
            for chosenSE in ses:
                if chosenSE in possibleSEs:
                    tasks.append((chosenSE, lfns))
                    existingCount[targetSite] = existingCount.setdefault(targetSite, 0) + len(lfns)
        return S_OK(tasks)

    def _getShares(self, shareType, normalise=False):
        """Takes share from the CS, eventually normalize them"""
        res = gConfig.getOptionsDict("/Resources/Shares/%s" % shareType)
        if not res["OK"]:
            return res
        if not res["Value"]:
            return S_ERROR("/Resources/Shares/%s option contains no shares" % shareType)
        shares = res["Value"]
        for site, value in shares.items():
            shares[site] = float(value)
        if normalise:
            shares = self.util._normaliseShares(shares)  # pylint: disable=protected-access
        if not shares:
            return S_ERROR("No non-zero shares defined")
        return S_OK(shares)

    def _getNextSite(self, existingCount, targetShares, candidates=None):
        """
        Selects the most suitable site, i.e. further to targetShares in existingCount
        """
        if candidates is None:
            candidates = targetShares
        # normalise the existing counts
        existingShares = self.util._normaliseShares(existingCount)  # pylint: disable=protected-access
        # then fill the missing share values to 0
        for site in targetShares:
            existingShares.setdefault(site, 0.0)
        # determine which site is farthest from its share
        chosenSite = ""
        minShareShortFall = -float("inf")
        for site, targetShare in targetShares.items():
            if site not in candidates or not targetShare:
                continue
            existingShare = existingShares[site]
            shareShortFall = targetShare - existingShare
            if shareShortFall > minShareShortFall:
                minShareShortFall = shareShortFall
                chosenSite = site
        return S_OK(chosenSite)

    @classmethod
    def _getSiteForSE(cls, se):
        """Get site name for the given SE"""
        result = getSitesForSE(se)
        if not result["OK"]:
            return result
        if result["Value"]:
            return S_OK(result["Value"][0])
        return S_OK("")

    @classmethod
    def _getSitesForSEs(cls, seList):
        """Get all the sites for the given SE list"""
        sites = []
        for se in seList:
            result = getSitesForSE(se)
            if result["OK"]:
                sites += result["Value"]
        return sites



""" Command module contains the base "Command" class, base for all RSS commands

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC import gLogger, S_OK


class Command(object):
    """
    The Command class is a simple base class for all the commands
    for interacting with the clients
    """

    def __init__(self, args=None, clients=None):

        self.apis = (1 and clients) or {}
        self.masterMode = False
        self.onlyCache = False
        self.metrics = {"failed": []}

        self.args = {"onlyCache": False}
        _args = (1 and args) or {}
        self.args.update(_args)
        self.log = gLogger.getSubLogger(self.__class__.__name__)

    def doNew(self, masterParams=None):
        """To be extended/replaced by real commands"""
        return S_OK((self.args, masterParams))

    def doCache(self):
        """To be extended/replaced by real commands"""
        return S_OK(self.args)

    def doMaster(self):
        """To be extended/replaced by real commands"""
        return S_OK(self.metrics)

    def doCommand(self):
        """Real commands don't need to extende/replace this method, which is called by clients.

        What is done here is the following:
        if self.masterMode is set to True, then the "doMaster()" method is called.
        if not, then the doCache() method is called, and if this returns an object this is returned,
        and otherwise the "doNew" method is called.
        """

        if self.masterMode:
            self.log.verbose("doMaster")
            return self.returnSObj(self.doMaster())

        self.log.verbose("doCache")
        result = self.doCache()
        if not result["OK"]:
            return self.returnERROR(result)
        # We may be interested on running the commands only from the cache,
        # without requesting new values.
        if result["Value"] or self.args["onlyCache"]:
            return result

        self.log.verbose("doNew")
        return self.returnSObj(self.doNew())

    def returnERROR(self, s_obj):
        """
        Overwrites S_ERROR message with command name, much easier to debug
        """

        s_obj["Message"] = "%s %s" % (self.__class__.__name__, s_obj["Message"])
        return s_obj

    def returnSObj(self, s_obj):
        """
        Overwrites S_ERROR message with command name, much easier to debug
        """

        if s_obj["OK"]:
            return s_obj

        return self.returnERROR(s_obj)


################################################################################
# EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import time

from tornado import gen

# We change the import name otherwise sphinx tries
# to compile the tornado doc and fails
from tornado.ioloop import IOLoop as _IOLoop

from DIRAC.ConfigurationSystem.Client.ConfigurationData import gConfigurationData
from DIRAC.ConfigurationSystem.private.RefresherBase import RefresherBase
from DIRAC.FrameworkSystem.Client.Logger import gLogger


class TornadoRefresher(RefresherBase):
    """
    The refresher, modified for Tornado
    It's the same refresher, the only thing which change is
    that we are using the IOLoop instead of threads for background
    tasks, so it work with Tornado (HTTPS server).
    """

    def refreshConfigurationIfNeeded(self):
        """
        Trigger an automatic refresh, most of the time nothing happens because automaticUpdate is enabled.
        This function is called by gConfig.getValue most of the time.

        We disable pylint error because this class must be instanciated by a mixin to define the missing methods
        """
        if not self._refreshEnabled or self._automaticUpdate:  # pylint: disable=no-member
            return
        if not gConfigurationData.getServers() or not self._lastRefreshExpired():  # pylint: disable=no-member
            return
        self._lastUpdateTime = time.time()
        _IOLoop.current().run_in_executor(None, self._refresh)  # pylint: disable=no-member

    def autoRefreshAndPublish(self, sURL):
        """
        Start the autorefresh background task, called by ServiceInterface
        (the class behind the Configuration/Server handler)

        :param str sURL: URL of the configuration server
        """
        gLogger.debug("Setting configuration refresh as automatic")
        if not gConfigurationData.getAutoPublish():
            gLogger.debug("Slave server won't auto publish itself")
        if not gConfigurationData.getName():
            import DIRAC

            DIRAC.abort(10, "Missing configuration name!")
        self._url = sURL
        self._automaticUpdate = True

        # Tornado replacement solution to the classic thread
        # It start the method self.__refreshLoop on the next IOLoop iteration
        _IOLoop.current().spawn_callback(self.__refreshLoop)

    @gen.coroutine
    def __refreshLoop(self):
        """
        Trigger the autorefresh when configuration is expired

        This task must use Tornado utilities to avoid blocking the ioloop and
        potentialy deadlock the server.

        See http://www.tornadoweb.org/en/stable/guide/coroutines.html#looping
        for official documentation about this type of method.
        """
        while self._automaticUpdate:

            # This is the sleep from Tornado, like a sleep it wait some time
            # But this version is non-blocking, so IOLoop can continue execution
            yield gen.sleep(gConfigurationData.getPropagationTime())
            # Publish step is blocking so we have to run it in executor
            # If we are not doing it, when master try to ping we block the IOLoop
            yield _IOLoop.current().run_in_executor(None, self.__AutoRefresh)

    @gen.coroutine
    def __AutoRefresh(self):
        """
        Auto refresh the configuration
        We disable pylint error because this class must be instanciated
        by a mixin to define the methods.
        """
        if self._refreshEnabled:  # pylint: disable=no-member
            if not self._refreshAndPublish():  # pylint: disable=no-member
                gLogger.error("Can't refresh configuration from any source")

    def daemonize(self):
        """daemonize is probably not the best name because there is no daemon behind
        but we must keep it to the same interface of the DISET refresher"""
        _IOLoop.current().spawn_callback(self.__refreshLoop)



"""  StatesAccountingAgent sends periodically numbers of jobs in various states for various
     sites to the Monitoring system to create historical plots.

.. literalinclude:: ../ConfigTemplate.cfg
  :start-after: ##BEGIN StatesAccountingAgent
  :end-before: ##END
  :dedent: 2
  :caption: StatesAccountingAgent options
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.Core.Utilities import Time
from DIRAC.AccountingSystem.Client.Types.WMSHistory import WMSHistory
from DIRAC.AccountingSystem.Client.DataStoreClient import DataStoreClient
from DIRAC.MonitoringSystem.Client.MonitoringReporter import MonitoringReporter
from DIRAC.WorkloadManagementSystem.DB.JobDB import JobDB


class StatesAccountingAgent(AgentModule):
    """Agent that every 15 minutes will report
    to the AccountingDB (MySQL) or the Monitoring DB (ElasticSearch), or both,
    a snapshot of the JobDB.
    """

    __summaryKeyFieldsMapping = [
        "Status",
        "Site",
        "User",
        "UserGroup",
        "JobGroup",
        "JobType",
        "ApplicationStatus",
        "MinorStatus",
    ]
    __summaryDefinedFields = [("ApplicationStatus", "unset"), ("MinorStatus", "unset")]
    __summaryValueFieldsMapping = ["Jobs", "Reschedules"]
    __renameFieldsMapping = {"JobType": "JobSplitType"}

    def initialize(self):
        """Standard initialization"""
        # This agent will always loop every 15 minutes
        self.am_setOption("PollingTime", 900)

        self.backends = self.am_getOption("Backends", "Accounting").replace(" ", "").split(",")
        messageQueue = self.am_getOption("MessageQueue", "dirac.wmshistory")

        self.log.info("Committing to %s backend" % "and ".join(self.backends))

        self.datastores = {}  # For storing the clients to Accounting and Monitoring

        if "Accounting" in self.backends:
            self.datastores["Accounting"] = DataStoreClient(retryGraceTime=900)
        if "Monitoring" in self.backends:
            self.datastores["Monitoring"] = MonitoringReporter(
                monitoringType="WMSHistory", failoverQueueName=messageQueue
            )

        self.__jobDBFields = []
        for field in self.__summaryKeyFieldsMapping:
            if field == "User":
                field = "Owner"
            elif field == "UserGroup":
                field = "OwnerGroup"
            self.__jobDBFields.append(field)
        return S_OK()

    def execute(self):
        """Main execution method"""
        # Get the WMS Snapshot!
        result = JobDB().getSummarySnapshot(self.__jobDBFields)
        now = Time.dateTime()
        if not result["OK"]:
            self.log.error("Can't get the JobDB summary", "%s: won't commit at this cycle" % result["Message"])
            return S_ERROR()

        # Now we try to commit
        values = result["Value"][1]

        self.log.info("Start sending records")
        for record in values:
            record = record[1:]
            rD = {}
            for fV in self.__summaryDefinedFields:
                rD[fV[0]] = fV[1]
            for iP in range(len(self.__summaryKeyFieldsMapping)):
                fieldName = self.__summaryKeyFieldsMapping[iP]
                rD[self.__renameFieldsMapping.get(fieldName, fieldName)] = record[iP]
            record = record[len(self.__summaryKeyFieldsMapping) :]
            for iP in range(len(self.__summaryValueFieldsMapping)):
                rD[self.__summaryValueFieldsMapping[iP]] = int(record[iP])

            for backend in self.datastores:
                if backend.lower() == "monitoring":
                    rD["timestamp"] = int(Time.toEpoch(now))
                    self.datastores["Monitoring"].addRecord(rD)

                elif backend.lower() == "accounting":
                    acWMS = WMSHistory()
                    acWMS.setStartTime(now)
                    acWMS.setEndTime(now)
                    acWMS.setValuesFromDict(rD)
                    retVal = acWMS.checkValues()
                    if not retVal["OK"]:
                        self.log.error("Invalid accounting record ", "%s -> %s" % (retVal["Message"], rD))
                    else:
                        self.datastores["Accounting"].addRegister(acWMS)

        for backend, datastore in self.datastores.items():
            self.log.info("Committing to %s backend" % backend)
            result = datastore.commit()
            if not result["OK"]:
                self.log.error("Couldn't commit WMS history to %s" % backend, result["Message"])
                return S_ERROR()
            self.log.verbose("Done committing to %s backend" % backend)

        return S_OK()



""" pytest for WorkflowTasks
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# pylint: disable=protected-access,missing-docstring,invalid-name
import json
import six
import sys
from mock import MagicMock

from pytest import mark

parametrize = mark.parametrize

from hypothesis import given, settings
from hypothesis.strategies import (
    composite,
    integers,
    lists,
    text,
    dictionaries,
    from_regex,
)
from string import ascii_letters, digits


from DIRAC.Core.Utilities.JEncode import encode


from DIRAC.TransformationSystem.Client.BodyPlugin.DummyBody import DummyBody
from DIRAC.TransformationSystem.Client.RequestTasks import RequestTasks


@composite
def taskStrategy(draw):
    """Generate a strategy that returns a task dictionary"""
    transformationID = draw(integers(min_value=1))
    targetSE = ",".join(draw(lists(text(ascii_letters, min_size=5, max_size=10), min_size=1, max_size=3)))
    inputData = draw(lists(from_regex("(/[a-z]+)+", fullmatch=True), min_size=1, max_size=10))

    return {"TransformationID": transformationID, "TargetSE": targetSE, "InputData": inputData}


def taskDictStrategy():
    return dictionaries(
        integers(min_value=1), taskStrategy(), min_size=1, max_size=10  # pylint: disable=no-value-for-parameter
    )


mockTransClient = MagicMock()
mockReqClient = MagicMock()
mockReqValidator = MagicMock()


reqTasks = RequestTasks(
    transClient=mockTransClient,
    requestClient=mockReqClient,
    requestValidator=mockReqValidator,
)


@parametrize(
    "transBody",
    [
        "removal;RemoveFile",  # Transformation to remove files
        "removal;RemoveReplica",  # Transformation to remove Replicas
        "anything;ReplicateAndRegister",  # Transformation to replicate and register, first parameter is useless
        "",  # if no Body, we expect replicateAndRegister
    ],
)
@mark.slow
@settings(max_examples=50)
@given(
    owner=text(ascii_letters + "-_" + digits, min_size=1),
    taskDict=taskDictStrategy(),
)
def test_prepareSingleOperationsBody(transBody, owner, taskDict):
    """Test different bodies that should be routed through the
    singleOperationBody method.
    """

    # keep the number of tasks for later
    originalNbOfTasks = len(taskDict)

    # Make up the DN and the group
    ownerDN = "DN_" + owner
    ownerGroup = "group_" + owner

    res = reqTasks.prepareTransformationTasks(transBody, taskDict, owner=owner, ownerGroup=ownerGroup, ownerDN=ownerDN)

    assert res["OK"], res

    # prepareTransformationTasks can pop tasks if a problem occurs,
    # so check that this did not happen
    assert len(res["Value"]) == originalNbOfTasks

    for _taskID, task in taskDict.items():

        req = task.get("TaskObject")

        # Checks whether we got a Request assigned
        assert req

        # Check that the attributes of the request are what
        # we expect them to be
        assert req.OwnerDN == ownerDN
        assert req.OwnerGroup == ownerGroup

        # Make sure we only have one operation
        assert len(req) == 1

        ops = req[0]

        # Check the operation type
        # The operation type is either given as second parameter of the body
        # or if not, it is ReplicateAndRegister

        expectedOpsType = transBody.split(";")[-1] if transBody else "ReplicateAndRegister"
        assert ops.Type == expectedOpsType

        # Check that the targetSE is set correctly
        assert ops.TargetSE == task["TargetSE"]

        # Make sure we have one file per LFN in the task
        assert len(ops) == len(task["InputData"])

        # Checks that there is one file per LFN
        assert set([f.LFN for f in ops]) == set(task["InputData"])


@parametrize(
    "transBody",
    [
        [
            ("ReplicateAndRegister", {"TargetSE": "BAR-SRM"}),
        ],
        [
            ("ReplicateAndRegister", {"TargetSE": "TASK:TargetSE"}),
        ],
        [
            ("ReplicateAndRegister", {"SourceSE": "FOO-SRM", "TargetSE": "BAR-SRM"}),
            ("RemoveReplica", {"TargetSE": "FOO-SRM"}),
        ],
        [
            ("ReplicateAndRegister", {"SourceSE": "FOO-SRM", "TargetSE": "TASK:TargetSE"}),
            ("RemoveReplica", {"TargetSE": "FOO-SRM"}),
        ],
    ],
    ids=[
        "Single operation, no substitution",
        "Single operation, with substitution",
        "Multiple operations, no substitution",
        "Multiple operations, with substitution",
    ],
)
@mark.slow
@settings(max_examples=50)
@given(
    owner=text(ascii_letters + "-_" + digits, min_size=1),
    taskDict=taskDictStrategy(),
)
def test_prepareMultiOperationsBody(transBody, owner, taskDict):
    """Test different bodies that should be routed through the
    multiOperationsBody method.
    """

    # keep the number of tasks for later
    originalNbOfTasks = len(taskDict)

    # Make up the DN and the group
    ownerDN = "DN_" + owner
    ownerGroup = "group_" + owner

    res = reqTasks.prepareTransformationTasks(
        json.dumps(transBody), taskDict, owner=owner, ownerGroup=ownerGroup, ownerDN=ownerDN
    )

    assert res["OK"], res

    # prepareTransformationTasks can pop tasks if a problem occurs,
    # so check that this did not happen
    assert len(res["Value"]) == originalNbOfTasks

    for _taskID, task in taskDict.items():

        req = task.get("TaskObject")

        # Checks whether we got a Request assigned
        assert req

        # Check that the attributes of the request are what
        # we expect them to be
        assert req.OwnerDN == ownerDN
        assert req.OwnerGroup == ownerGroup

        # Make sure we have as many operations as tuple in the body
        assert len(req) == len(transBody)

        # Loop over each operation
        # to check their attributes
        for opsID, ops in enumerate(req):

            expectedOpsType, expectedOpsAttributes = transBody[opsID]

            # Compare the operation type with what we want
            assert ops.Type == expectedOpsType

            # Check the operation attributes one after the other
            for opsAttr, opsVal in expectedOpsAttributes.items():

                # If the expected value starts with 'TASK:'
                # we should make the substitution with whatever is in
                # the taskDict.
                # So it should be different
                if opsVal.startswith("TASK:"):
                    assert getattr(ops, opsAttr) != opsVal
                # Otherwise, make sure it matches
                else:
                    assert getattr(ops, opsAttr) == opsVal

                # Make sure we have one file per LFN in the task
                assert len(ops) == len(task["InputData"])

                # Checks that there is one file per LFN
                assert set([f.LFN for f in ops]) == set(task["InputData"])


@parametrize(
    "transBody",
    [
        # We request a key that does not exist in the taskDict
        [
            ("ReplicateAndRegister", {"TargetSE": "TASK:NotInTaskDict"}),
        ],
    ],
    ids=[
        "Non existing substitution",
    ],
)
@mark.slow
@settings(max_examples=50)
@given(
    owner=text(ascii_letters + "-_" + digits, min_size=1),
    taskDict=taskDictStrategy(),
)
def test_prepareProblematicMultiOperationsBody(transBody, owner, taskDict):
    """Test different bodies that should be routed through the
    multiOperationBody method, but that have a problem
    """

    # keep the number of tasks for later
    originalNbOfTasks = len(taskDict)

    # Make up the DN and the group
    ownerDN = "DN_" + owner
    ownerGroup = "group_" + owner

    res = reqTasks.prepareTransformationTasks(
        json.dumps(transBody), taskDict, owner=owner, ownerGroup=ownerGroup, ownerDN=ownerDN
    )

    assert res["OK"], res

    # prepareTransformationTasks pop tasks if a problem occurs,
    # so make sure it happened
    assert len(res["Value"]) != originalNbOfTasks

    # Check that other tasks are fine
    # Note: currently, we will never enter this loop as
    # the transbody are buggy, so all the tasks should be removed.
    # I just prepare the future :-)

    for _taskID, task in taskDict.items():

        req = task.get("TaskObject")

        # Checks whether we got a Request assigned
        assert req

        # Check that the attributes of the request are what
        # we expect them to be
        assert req.OwnerDN == ownerDN
        assert req.OwnerGroup == ownerGroup

        # Make sure we have as many operations as tuple in the body
        assert len(req) == len(transBody)

        # Loop over each operation
        # to check their attributes
        for opsID, ops in enumerate(req):

            expectedOpsType, expectedOpsAttributes = transBody[opsID]

            # Compare the operation type with what we want
            assert ops.Type == expectedOpsType

            # Check the operation attributes one after the other
            for opsAttr, opsVal in expectedOpsAttributes.items():

                # If the expected value starts with 'TASK:'
                # we should make the substitution with whatever is in
                # the taskDict.
                # So it should be different
                if opsVal.startswith("TASK:"):
                    assert getattr(ops, opsAttr) != opsVal
                # Otherwise, make sure it matches
                else:
                    # Check that the targetSE is set correctly
                    assert getattr(ops, opsAttr) == opsVal

                # Make sure we have one file per LFN in the task
                assert len(ops) == len(task["InputData"])

                # Checks that there is one file per LFN
                assert set([f.LFN for f in ops]) == set(task["InputData"])


@mark.slow
@settings(max_examples=50, deadline=500)
@given(
    taskDict=taskDictStrategy(),
    pluginFactor=integers(),
)
def test_complexBodyPlugin(taskDict, pluginFactor):
    """This test makes sure that we can load the BodyPlugin objects"""

    transBody = DummyBody(factor=pluginFactor)

    # keep the number of tasks for later
    originalNbOfTasks = len(taskDict)

    # Make up the DN and the group
    ownerDN = "DN_owner"
    ownerGroup = "group_owner"

    res = reqTasks.prepareTransformationTasks(
        encode(transBody), taskDict, owner="owner", ownerGroup=ownerGroup, ownerDN=ownerDN
    )

    assert res["OK"], res

    # prepareTransformationTasks can pop tasks if a problem occurs,
    # so check that this did not happen
    assert len(res["Value"]) == originalNbOfTasks

    for _taskID, task in taskDict.items():

        req = task.get("TaskObject")

        # Checks whether we got a Request assigned
        assert req

        # Check that the attributes of the request are what
        # we expect them to be
        assert req.OwnerDN == ownerDN
        assert req.OwnerGroup == ownerGroup

        # DummyBody only creates a single operation.
        # It should be a forward diset, and the
        # argument should be the number of files in the task
        # multiplied by the pluginParam

        assert len(req) == 1

        ops = req[0]
        assert ops.Type == "ForwardDISET"
        assert json.loads(ops.Arguments) == pluginFactor * len(task["InputData"])



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# Just creating some processes, for tests

import time
from multiprocessing import Pool


def f(n):
    time.sleep(n)


p = Pool(3)
p.map(f, [3, 4, 5])



""" PolicyBase

  The Policy class is a simple base class for all the policies.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC.ResourceStatusSystem.Command.Command import Command

__RCSID__ = "$Id$"


class PolicyBase(object):
    """Base class for all the policies. Do not instantiate directly.
    To use, you should call `setCommand` on the real policy instance.
    """

    def __init__(self):
        """Constructor"""

        self.command = Command()
        self.result = {}

    def setCommand(self, policyCommand):
        """
        Set `self.command`.

        :params:
          :attr:`commandIn`: a command object
        """
        if policyCommand is not None:
            self.command = policyCommand

    def evaluate(self):
        """
        Before use, call `setCommand`.

        Invoking `super(PolicyCLASS, self).evaluate` will invoke
        the command (if necessary) as it is provided and returns the results.
        """

        commandResult = self.command.doCommand()
        return self._evaluate(commandResult)

    @staticmethod
    def _evaluate(commandResult):
        """
        Method that will do the real processing of the policy, it has to be extended
        on the real policies.
        """

        return commandResult



"""RabbitMQAdmin module serves for the management of the internal RabbitMQ
   users database. It uses rabbitmqctl command. Only the user with the right
   permissions can execute those commands.
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
import re
from DIRAC import S_OK, S_ERROR
import errno
from DIRAC.Core.Utilities import Subprocess


def executeRabbitmqctl(arg, *argv):
    """Executes RabbitMQ administration command.
      It uses rabbitmqctl command line interface.
      For every command the -q argument ("quit mode")
      is used, since in some cases the output must be processed,
      so we don't want any additional informations printed.

    Args:
      arg(str): command recognized by the rabbitmqctl.
      argv: optional list of string parameters.

    :rtype: S_OK or S_ERROR
    :type argv: python:list

    """
    command = ["sudo", "/usr/sbin/rabbitmqctl", "-q", arg] + list(argv)
    timeOut = 30
    result = Subprocess.systemCall(timeout=timeOut, cmdSeq=command)
    if not result["OK"]:
        return S_ERROR(errno.EPERM, "%r failed to launch" % command)
    errorcode, cmd_out, cmd_err = result["Value"]
    if errorcode:
        # No idea what errno code should be used here.
        # Maybe we should define some specific for rabbitmqctl
        return S_ERROR(
            errno.EPERM, "%r failed, status code: %s stdout: %r stderr: %r" % (command, errorcode, cmd_out, cmd_err)
        )
    return S_OK(cmd_out)


def addUserWithoutPassword(user):
    """Adds user to the internal RabbitMQ database
    and clears its password.
    This should be done for all users, that
    will be using SSL authentication. They do not
    need any password.
    """
    ret = addUser(user)
    if not ret["OK"]:
        return ret
    return clearUserPassword(user)


def addUser(user, password="password"):
    """Adds user to the internal RabbitMQ database
    Function also sets user password.
    User still cannot access to any resources, without
    having permissions set.
    """
    return executeRabbitmqctl("add_user", user, password)


def deleteUser(user):
    """Removes the user from the internal RabbitMQ database."""
    return executeRabbitmqctl("delete_user", user)


def getAllUsers():
    """Returns all existing users in the internal RabbitMQ database.

    :returns: S_OK with a list of all users
    :rtype: S_OK
    """
    ret = executeRabbitmqctl("list_users")
    if not ret["OK"]:
        return ret
    users = ret["Value"]
    users = users.split("\n")
    # the rabbitMQ user list is given in the format:
    # user_name [usr_tag]
    # I remove [usr_tag] part.
    # Also only non-empty users are proceeded further.
    # Empty users can appear, cause every new line was
    # treated as a new user.
    users = [re.sub(r"\\t\[\w*\]$", "", u) for u in users if u]
    return S_OK(users)


def setUserPermission(user):
    return executeRabbitmqctl("set_permissions", "-p", "/", user, '".*"', '".*"', '".*"')


def clearUserPassword(user):
    """Clears users password for the internal RabbitMQ
    database. User with no password cannot enter
    the RabbitMQ website interface but still can
    connect via SSL if given permission.
    """
    return executeRabbitmqctl("clear_password", user)


def setUsersPermissions(users):
    successful = {}
    failed = {}
    for u in users:
        ret = setUserPermission(u)
        if ret["OK"]:
            successful[u] = ret["Value"]
        else:
            print("Problem with permissions:%s" % ret["Message"])
            failed[u] = "Permission not set because of:%s" % ret["Message"]
    return S_OK({"Successful": successful, "Failed": failed})


def addUsersWithoutPasswords(users):
    successful = {}
    failed = {}
    for u in users:
        ret = addUserWithoutPassword(u)
        if ret["OK"]:
            successful[u] = ret["Value"]
        else:
            print("Problem with adding user:%s" % ret["Message"])
            failed[u] = "User not added"
    return S_OK({"Successful": successful, "Failed": failed})


def addUsers(users):
    """Adds users to the RabbitMQ internal database."""
    successful = {}
    failed = {}
    for u in users:
        ret = addUser(u)
        if ret["OK"]:
            successful[u] = ret["Value"]
        else:
            print("Problem with adding user:%s" % ret["Message"])
            failed[u] = "User not added"
    return S_OK({"Successful": successful, "Failed": failed})


def deleteUsers(users):
    """Deletes users from the RabbitMQ internal database."""
    successful = {}
    failed = {}
    for u in users:
        ret = deleteUser(u)
        if ret["OK"]:
            successful[u] = ret["Value"]
        else:
            print("Problem with adding user:%s" % ret["Message"])
            failed[u] = "User not added"
    return S_OK({"Successful": successful, "Failed": failed})



""" Base corrector for the group and ingroup shares
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC import S_OK, S_ERROR

__RCSID__ = "$Id$"


class BaseCorrector(object):
    def __init__(self, opsHelper, baseCSPath, group):
        self.__opsHelper = opsHelper
        self.__baseCSPath = baseCSPath
        self.__group = group

    def initialize(self):
        return S_OK()

    def getCSOption(self, opName, defValue=None):
        return self.__opsHelper.getValue("%s/%s" % (self.__baseCSPath, opName), defValue)

    def getiCSOptions(self, opName=""):
        return self.__opsHelper.getSections("%s/%s" % (self.__baseCSPath, opName))

    def getCSSections(self, secName=""):
        return self.__opsHelper.getSections("%s/%s" % (self.__baseCSPath, secName))

    def getGroup(self):
        return self.__group

    def updateHistoryKnowledge(self):
        return S_OK()

    def applyCorrection(self, _entitiesExpectedShare):
        return S_ERROR("applyCorrection function has not been implemented")



""" ProxyManagerClient has the function to "talk" to the ProxyManager service
    (:mod:`~DIRAC.FrameworkSystem.Service.ProxyManagerHandler`).
    This inherits the DIRAC base Client for direct execution of server functionality.
    Client also contain caching of the requested proxy information.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import six
import os
import datetime

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.ConfigurationSystem.Client.Helpers import Registry
from DIRAC.Core.Utilities import ThreadSafe, DIRACSingleton
from DIRAC.Core.Utilities.DictCache import DictCache
from DIRAC.Core.Security.ProxyFile import multiProxyArgument, deleteMultiProxy
from DIRAC.Core.Security.X509Chain import X509Chain  # pylint: disable=import-error
from DIRAC.Core.Security.X509Request import X509Request  # pylint: disable=import-error
from DIRAC.Core.Security.VOMS import VOMS
from DIRAC.Core.Security import Locations
from DIRAC.Core.DISET.RPCClient import RPCClient

__RCSID__ = "$Id$"

gUsersSync = ThreadSafe.Synchronizer()
gProxiesSync = ThreadSafe.Synchronizer()
gVOMSProxiesSync = ThreadSafe.Synchronizer()


@six.add_metaclass(DIRACSingleton.DIRACSingleton)
class ProxyManagerClient(object):
    def __init__(self):
        self.__usersCache = DictCache()
        self.__proxiesCache = DictCache()
        self.__vomsProxiesCache = DictCache()
        self.__pilotProxiesCache = DictCache()
        self.__filesCache = DictCache(self.__deleteTemporalFile)

    def __deleteTemporalFile(self, filename):
        """Delete temporal file

        :param str filename: path to file
        """
        try:
            os.remove(filename)
        except Exception:
            pass

    def clearCaches(self):
        """Clear caches"""
        self.__usersCache.purgeAll()
        self.__proxiesCache.purgeAll()
        self.__vomsProxiesCache.purgeAll()
        self.__pilotProxiesCache.purgeAll()

    def __getSecondsLeftToExpiration(self, expiration, utc=True):
        """Get time left to expiration in a seconds

        :param datetime expiration:
        :param boolean utc: time in utc

        :return: datetime
        """
        if utc:
            td = expiration - datetime.datetime.utcnow()
        else:
            td = expiration - datetime.datetime.now()
        return td.days * 86400 + td.seconds

    def __refreshUserCache(self, validSeconds=0):
        """Refresh user cache

        :param int validSeconds: required seconds the proxy is valid for

        :return: S_OK()/S_ERROR()
        """
        rpcClient = RPCClient("Framework/ProxyManager", timeout=120)
        retVal = rpcClient.getRegisteredUsers(validSeconds)
        if not retVal["OK"]:
            return retVal
        data = retVal["Value"]
        # Update the cache
        for record in data:
            cacheKey = (record["DN"], record["group"])
            self.__usersCache.add(cacheKey, self.__getSecondsLeftToExpiration(record["expirationtime"]), record)
        return S_OK()

    @gUsersSync
    def userHasProxy(self, userDN, userGroup, validSeconds=0):
        """Check if a user(DN-group) has a proxy in the proxy management
        Updates internal cache if needed to minimize queries to the service

        :param str userDN: user DN
        :param str userGroup: user group
        :param int validSeconds: proxy valid time in a seconds

        :return: S_OK()/S_ERROR()
        """

        # For backward compatibility reasons with versions prior to v7r1
        # we need to check for proxy with a group
        # AND for groupless proxy even if not specified

        cacheKeys = ((userDN, userGroup), (userDN, ""))
        for cacheKey in cacheKeys:
            if self.__usersCache.exists(cacheKey, validSeconds):
                return S_OK(True)

        # Get list of users from the DB with proxys at least 300 seconds
        gLogger.verbose("Updating list of users in proxy management")
        retVal = self.__refreshUserCache(validSeconds)
        if not retVal["OK"]:
            return retVal

        for cacheKey in cacheKeys:
            if self.__usersCache.exists(cacheKey, validSeconds):
                return S_OK(True)

        return S_OK(False)

    @gUsersSync
    def getUserPersistence(self, userDN, userGroup, validSeconds=0):
        """Check if a user(DN-group) has a proxy in the proxy management
        Updates internal cache if needed to minimize queries to the service

        :param str userDN: user DN
        :param str userGroup: user group
        :param int validSeconds: proxy valid time in a seconds

        :return: S_OK()/S_ERROR()
        """
        cacheKey = (userDN, userGroup)
        userData = self.__usersCache.get(cacheKey, validSeconds)
        if userData:
            if userData["persistent"]:
                return S_OK(True)
        # Get list of users from the DB with proxys at least 300 seconds
        gLogger.verbose("Updating list of users in proxy management")
        retVal = self.__refreshUserCache(validSeconds)
        if not retVal["OK"]:
            return retVal
        userData = self.__usersCache.get(cacheKey, validSeconds)
        if userData:
            return S_OK(userData["persistent"])
        return S_OK(False)

    def setPersistency(self, userDN, userGroup, persistent):
        """Set the persistency for user/group

        :param str userDN: user DN
        :param str userGroup: user group
        :param boolean persistent: presistent flag

        :return: S_OK()/S_ERROR()
        """
        # Hack to ensure bool in the rpc call
        persistentFlag = True
        if not persistent:
            persistentFlag = False
        rpcClient = RPCClient("Framework/ProxyManager", timeout=120)
        retVal = rpcClient.setPersistency(userDN, userGroup, persistentFlag)
        if not retVal["OK"]:
            return retVal
        # Update internal persistency cache
        cacheKey = (userDN, userGroup)
        record = self.__usersCache.get(cacheKey, 0)
        if record:
            record["persistent"] = persistentFlag
            self.__usersCache.add(cacheKey, self.__getSecondsLeftToExpiration(record["expirationtime"]), record)
        return retVal

    def uploadProxy(self, proxy=None, restrictLifeTime=0, rfcIfPossible=False):
        """Upload a proxy to the proxy management service using delegation

        :param X509Chain proxy: proxy as a chain
        :param int restrictLifeTime: proxy live time in a seconds
        :param boolean rfcIfPossible: make rfc proxy if possible

        :return: S_OK(dict)/S_ERROR() -- dict contain proxies
        """
        # Discover proxy location
        if isinstance(proxy, X509Chain):
            chain = proxy
            proxyLocation = ""
        else:
            if not proxy:
                proxyLocation = Locations.getProxyLocation()
                if not proxyLocation:
                    return S_ERROR("Can't find a valid proxy")
            elif isinstance(proxy, six.string_types):
                proxyLocation = proxy
            else:
                return S_ERROR("Can't find a valid proxy")
            chain = X509Chain()
            result = chain.loadProxyFromFile(proxyLocation)
            if not result["OK"]:
                return S_ERROR("Can't load %s: %s " % (proxyLocation, result["Message"]))

        # Make sure it's valid
        if chain.hasExpired().get("Value"):
            return S_ERROR("Proxy %s has expired" % proxyLocation)
        if chain.getDIRACGroup(ignoreDefault=True).get("Value") or chain.isVOMS().get("Value"):
            return S_ERROR("Cannot upload proxy with DIRAC group or VOMS extensions")

        rpcClient = RPCClient("Framework/ProxyManager", timeout=120)
        # Get a delegation request
        result = rpcClient.requestDelegationUpload(chain.getRemainingSecs()["Value"])
        if not result["OK"]:
            return result
        reqDict = result["Value"]
        # Generate delegated chain
        chainLifeTime = chain.getRemainingSecs()["Value"] - 60
        if restrictLifeTime and restrictLifeTime < chainLifeTime:
            chainLifeTime = restrictLifeTime
        retVal = chain.generateChainFromRequestString(reqDict["request"], lifetime=chainLifeTime, rfc=rfcIfPossible)
        if not retVal["OK"]:
            return retVal
        # Upload!
        result = rpcClient.completeDelegationUpload(reqDict["id"], retVal["Value"])
        if not result["OK"]:
            return result
        return S_OK(result.get("proxies") or result["Value"])

    @gProxiesSync
    def downloadProxy(
        self, userDN, userGroup, limited=False, requiredTimeLeft=1200, cacheTime=14400, proxyToConnect=None, token=None
    ):
        """Get a proxy Chain from the proxy management

        :param str userDN: user DN
        :param str userGroup: user group
        :param boolean limited: if need limited proxy
        :param int requiredTimeLeft: required proxy live time in a seconds
        :param int cacheTime: store in a cache time in a seconds
        :param X509Chain proxyToConnect: proxy as a chain
        :param str token: valid token to get a proxy

        :return: S_OK(X509Chain)/S_ERROR()
        """
        cacheKey = (userDN, userGroup)
        if self.__proxiesCache.exists(cacheKey, requiredTimeLeft):
            return S_OK(self.__proxiesCache.get(cacheKey))
        req = X509Request()
        req.generateProxyRequest(limited=limited)
        if proxyToConnect:
            rpcClient = RPCClient("Framework/ProxyManager", proxyChain=proxyToConnect, timeout=120)
        else:
            rpcClient = RPCClient("Framework/ProxyManager", timeout=120)
        if token:
            retVal = rpcClient.getProxyWithToken(
                userDN, userGroup, req.dumpRequest()["Value"], int(cacheTime + requiredTimeLeft), token
            )
        else:
            retVal = rpcClient.getProxy(
                userDN, userGroup, req.dumpRequest()["Value"], int(cacheTime + requiredTimeLeft)
            )
        if not retVal["OK"]:
            return retVal
        chain = X509Chain(keyObj=req.getPKey())
        retVal = chain.loadChainFromString(retVal["Value"])
        if not retVal["OK"]:
            return retVal
        self.__proxiesCache.add(cacheKey, chain.getRemainingSecs()["Value"], chain)
        return S_OK(chain)

    def downloadProxyToFile(
        self,
        userDN,
        userGroup,
        limited=False,
        requiredTimeLeft=1200,
        cacheTime=14400,
        filePath=None,
        proxyToConnect=None,
        token=None,
    ):
        """Get a proxy Chain from the proxy management and write it to file

        :param str userDN: user DN
        :param str userGroup: user group
        :param boolean limited: if need limited proxy
        :param int requiredTimeLeft: required proxy live time in a seconds
        :param int cacheTime: store in a cache time in a seconds
        :param str filePath: path to save proxy
        :param X509Chain proxyToConnect: proxy as a chain
        :param str token: valid token to get a proxy

        :return: S_OK(X509Chain)/S_ERROR()
        """
        retVal = self.downloadProxy(userDN, userGroup, limited, requiredTimeLeft, cacheTime, proxyToConnect, token)
        if not retVal["OK"]:
            return retVal
        chain = retVal["Value"]
        retVal = self.dumpProxyToFile(chain, filePath)
        if not retVal["OK"]:
            return retVal
        retVal["chain"] = chain
        return retVal

    @gVOMSProxiesSync
    def downloadVOMSProxy(
        self,
        userDN,
        userGroup,
        limited=False,
        requiredTimeLeft=1200,
        cacheTime=14400,
        requiredVOMSAttribute=None,
        proxyToConnect=None,
        token=None,
    ):
        """Download a proxy if needed and transform it into a VOMS one

        :param str userDN: user DN
        :param str userGroup: user group
        :param boolean limited: if need limited proxy
        :param int requiredTimeLeft: required proxy live time in a seconds
        :param int cacheTime: store in a cache time in a seconds
        :param str requiredVOMSAttribute: VOMS attr to add to the proxy
        :param X509Chain proxyToConnect: proxy as a chain
        :param str token: valid token to get a proxy

        :return: S_OK(X509Chain)/S_ERROR()
        """
        cacheKey = (userDN, userGroup, requiredVOMSAttribute, limited)
        if self.__vomsProxiesCache.exists(cacheKey, requiredTimeLeft):
            return S_OK(self.__vomsProxiesCache.get(cacheKey))
        req = X509Request()
        req.generateProxyRequest(limited=limited)
        if proxyToConnect:
            rpcClient = RPCClient("Framework/ProxyManager", proxyChain=proxyToConnect, timeout=120)
        else:
            rpcClient = RPCClient("Framework/ProxyManager", timeout=120)
        if token:
            retVal = rpcClient.getVOMSProxyWithToken(
                userDN,
                userGroup,
                req.dumpRequest()["Value"],
                int(cacheTime + requiredTimeLeft),
                token,
                requiredVOMSAttribute,
            )

        else:
            retVal = rpcClient.getVOMSProxy(
                userDN, userGroup, req.dumpRequest()["Value"], int(cacheTime + requiredTimeLeft), requiredVOMSAttribute
            )
        if not retVal["OK"]:
            return retVal
        chain = X509Chain(keyObj=req.getPKey())
        retVal = chain.loadChainFromString(retVal["Value"])
        if not retVal["OK"]:
            return retVal
        self.__vomsProxiesCache.add(cacheKey, chain.getRemainingSecs()["Value"], chain)
        return S_OK(chain)

    def downloadVOMSProxyToFile(
        self,
        userDN,
        userGroup,
        limited=False,
        requiredTimeLeft=1200,
        cacheTime=14400,
        requiredVOMSAttribute=None,
        filePath=None,
        proxyToConnect=None,
        token=None,
    ):
        """Download a proxy if needed, transform it into a VOMS one and write it to file

        :param str userDN: user DN
        :param str userGroup: user group
        :param boolean limited: if need limited proxy
        :param int requiredTimeLeft: required proxy live time in a seconds
        :param int cacheTime: store in a cache time in a seconds
        :param str requiredVOMSAttribute: VOMS attr to add to the proxy
        :param str filePath: path to save proxy
        :param X509Chain proxyToConnect: proxy as a chain
        :param str token: valid token to get a proxy

        :return: S_OK(X509Chain)/S_ERROR()
        """
        retVal = self.downloadVOMSProxy(
            userDN, userGroup, limited, requiredTimeLeft, cacheTime, requiredVOMSAttribute, proxyToConnect, token
        )
        if not retVal["OK"]:
            return retVal
        chain = retVal["Value"]
        retVal = self.dumpProxyToFile(chain, filePath)
        if not retVal["OK"]:
            return retVal
        retVal["chain"] = chain
        return retVal

    def getPilotProxyFromDIRACGroup(self, userDN, userGroup, requiredTimeLeft=43200, proxyToConnect=None):
        """Download a pilot proxy with VOMS extensions depending on the group

        :param str userDN: user DN
        :param str userGroup: user group
        :param int requiredTimeLeft: required proxy live time in seconds
        :param X509Chain proxyToConnect: proxy as a chain

        :return: S_OK(X509Chain)/S_ERROR()
        """
        # Assign VOMS attribute
        vomsAttr = Registry.getVOMSAttributeForGroup(userGroup)
        if not vomsAttr:
            gLogger.warn("No voms attribute assigned to group %s when requested pilot proxy" % userGroup)
            return self.downloadProxy(
                userDN, userGroup, limited=False, requiredTimeLeft=requiredTimeLeft, proxyToConnect=proxyToConnect
            )
        else:
            return self.downloadVOMSProxy(
                userDN,
                userGroup,
                limited=False,
                requiredTimeLeft=requiredTimeLeft,
                requiredVOMSAttribute=vomsAttr,
                proxyToConnect=proxyToConnect,
            )

    def getPilotProxyFromVOMSGroup(self, userDN, vomsAttr, requiredTimeLeft=43200, proxyToConnect=None):
        """Download a pilot proxy with VOMS extensions depending on the group

        :param str userDN: user DN
        :param str vomsAttr: VOMS attribute
        :param int requiredTimeLeft: required proxy live time in a seconds
        :param X509Chain proxyToConnect: proxy as a chain

        :return: S_OK(X509Chain)/S_ERROR()
        """
        groups = Registry.getGroupsWithVOMSAttribute(vomsAttr)
        if not groups:
            return S_ERROR("No group found that has %s as voms attrs" % vomsAttr)

        for userGroup in groups:
            result = self.downloadVOMSProxy(
                userDN,
                userGroup,
                limited=False,
                requiredTimeLeft=requiredTimeLeft,
                requiredVOMSAttribute=vomsAttr,
                proxyToConnect=proxyToConnect,
            )
            if result["OK"]:
                return result
        return result

    def getPayloadProxyFromDIRACGroup(self, userDN, userGroup, requiredTimeLeft, token=None, proxyToConnect=None):
        """Download a payload proxy with VOMS extensions depending on the group

        :param str userDN: user DN
        :param str userGroup: user group
        :param int requiredTimeLeft: required proxy live time in a seconds
        :param str token: valid token to get a proxy
        :param X509Chain proxyToConnect: proxy as a chain

        :return: S_OK(X509Chain)/S_ERROR()
        """
        # Assign VOMS attribute
        vomsAttr = Registry.getVOMSAttributeForGroup(userGroup)
        if not vomsAttr:
            gLogger.verbose("No voms attribute assigned to group %s when requested payload proxy" % userGroup)
            return self.downloadProxy(
                userDN,
                userGroup,
                limited=True,
                requiredTimeLeft=requiredTimeLeft,
                proxyToConnect=proxyToConnect,
                token=token,
            )
        else:
            return self.downloadVOMSProxy(
                userDN,
                userGroup,
                limited=True,
                requiredTimeLeft=requiredTimeLeft,
                requiredVOMSAttribute=vomsAttr,
                proxyToConnect=proxyToConnect,
                token=token,
            )

    def getPayloadProxyFromVOMSGroup(self, userDN, vomsAttr, token, requiredTimeLeft, proxyToConnect=None):
        """Download a payload proxy with VOMS extensions depending on the VOMS attr

        :param str userDN: user DN
        :param str vomsAttr: VOMS attribute
        :param str token: valid token to get a proxy
        :param int requiredTimeLeft: required proxy live time in a seconds
        :param X509Chain proxyToConnect: proxy as a chain

        :return: S_OK(X509Chain)/S_ERROR()
        """
        groups = Registry.getGroupsWithVOMSAttribute(vomsAttr)
        if not groups:
            return S_ERROR("No group found that has %s as voms attrs" % vomsAttr)
        userGroup = groups[0]

        return self.downloadVOMSProxy(
            userDN,
            userGroup,
            limited=True,
            requiredTimeLeft=requiredTimeLeft,
            requiredVOMSAttribute=vomsAttr,
            proxyToConnect=proxyToConnect,
            token=token,
        )

    def dumpProxyToFile(self, chain, destinationFile=None, requiredTimeLeft=600):
        """Dump a proxy to a file. It's cached so multiple calls won't generate extra files

        :param X509Chain chain: proxy as a chain
        :param str destinationFile: path to store proxy
        :param int requiredTimeLeft: required proxy live time in a seconds

        :return: S_OK(str)/S_ERROR()
        """
        result = chain.hash()
        if not result["OK"]:
            return result
        cHash = result["Value"]
        if self.__filesCache.exists(cHash, requiredTimeLeft):
            filepath = self.__filesCache.get(cHash)
            if filepath and os.path.isfile(filepath):
                return S_OK(filepath)
            self.__filesCache.delete(cHash)
        retVal = chain.dumpAllToFile(destinationFile)
        if not retVal["OK"]:
            return retVal
        filename = retVal["Value"]
        self.__filesCache.add(cHash, chain.getRemainingSecs()["Value"], filename)
        return S_OK(filename)

    def deleteGeneratedProxyFile(self, chain):
        """Delete a file generated by a dump

        :param X509Chain chain: proxy as a chain

        :return: S_OK()
        """
        self.__filesCache.delete(chain)
        return S_OK()

    def deleteProxyBundle(self, idList):
        """delete a list of id's

        :param list,tuple idList: list of identity numbers

        :return: S_OK(int)/S_ERROR()
        """
        rpcClient = RPCClient("Framework/ProxyManager", timeout=120)
        return rpcClient.deleteProxyBundle(idList)

    def requestToken(self, requesterDN, requesterGroup, numUses=1):
        """Request a number of tokens. usesList must be a list of integers and each integer is the number of uses a token
        must have

        :param str requesterDN: user DN
        :param str requesterGroup: user group
        :param int numUses: number of uses

        :return: S_OK(tuple)/S_ERROR() -- tuple contain token, number uses
        """
        rpcClient = RPCClient("Framework/ProxyManager", timeout=120)
        return rpcClient.generateToken(requesterDN, requesterGroup, numUses)

    def renewProxy(self, proxyToBeRenewed=None, minLifeTime=3600, newProxyLifeTime=43200, proxyToConnect=None):
        """Renew a proxy using the ProxyManager

        :param X509Chain proxyToBeRenewed: proxy to renew
        :param int minLifeTime: if proxy life time is less than this, renew. Skip otherwise
        :param int newProxyLifeTime: life time of new proxy
        :param X509Chain proxyToConnect: proxy to use for connecting to the service

        :return: S_OK(X509Chain)/S_ERROR()
        """
        retVal = multiProxyArgument(proxyToBeRenewed)
        if not retVal["Value"]:
            return retVal
        proxyToRenewDict = retVal["Value"]

        secs = proxyToRenewDict["chain"].getRemainingSecs()["Value"]
        if secs > minLifeTime:
            deleteMultiProxy(proxyToRenewDict)
            return S_OK()

        if not proxyToConnect:
            proxyToConnectDict = {"chain": False, "tempFile": False}
        else:
            retVal = multiProxyArgument(proxyToConnect)
            if not retVal["Value"]:
                deleteMultiProxy(proxyToRenewDict)
                return retVal
            proxyToConnectDict = retVal["Value"]

        userDN = proxyToRenewDict["chain"].getIssuerCert()["Value"].getSubjectDN()["Value"]
        retVal = proxyToRenewDict["chain"].getDIRACGroup()
        if not retVal["OK"]:
            deleteMultiProxy(proxyToRenewDict)
            deleteMultiProxy(proxyToConnectDict)
            return retVal
        userGroup = retVal["Value"]
        limited = proxyToRenewDict["chain"].isLimitedProxy()["Value"]

        voms = VOMS()
        retVal = voms.getVOMSAttributes(proxyToRenewDict["chain"])
        if not retVal["OK"]:
            deleteMultiProxy(proxyToRenewDict)
            deleteMultiProxy(proxyToConnectDict)
            return retVal
        vomsAttrs = retVal["Value"]
        if vomsAttrs:
            retVal = self.downloadVOMSProxy(
                userDN,
                userGroup,
                limited=limited,
                requiredTimeLeft=newProxyLifeTime,
                requiredVOMSAttribute=vomsAttrs[0],
                proxyToConnect=proxyToConnectDict["chain"],
            )
        else:
            retVal = self.downloadProxy(
                userDN,
                userGroup,
                limited=limited,
                requiredTimeLeft=newProxyLifeTime,
                proxyToConnect=proxyToConnectDict["chain"],
            )

        deleteMultiProxy(proxyToRenewDict)
        deleteMultiProxy(proxyToConnectDict)

        if not retVal["OK"]:
            return retVal

        chain = retVal["Value"]

        if not proxyToRenewDict["tempFile"]:
            return chain.dumpAllToFile(proxyToRenewDict["file"])

        return S_OK(chain)

    def getDBContents(self, condDict={}, sorting=[["UserDN", "DESC"]], start=0, limit=0):
        """Get the contents of the db

        :param dict condDict: search condition

        :return: S_OK(dict)/S_ERROR() -- dict contain fields, record list, total records
        """
        rpcClient = RPCClient("Framework/ProxyManager", timeout=120)
        return rpcClient.getContents(condDict, sorting, start, limit)

    def getVOMSAttributes(self, chain):
        """Get the voms attributes for a chain

        :param X509Chain chain: proxy as a chain

        :return: S_OK(str)/S_ERROR()
        """
        return VOMS().getVOMSAttributes(chain)

    def getUploadedProxyLifeTime(self, DN, group):
        """Get the remaining seconds for an uploaded proxy

        :param str DN: user DN
        :param str group: group

        :return: S_OK(int)/S_ERROR()
        """
        result = self.getDBContents({"UserDN": [DN], "UserGroup": [group]})
        if not result["OK"]:
            return result
        data = result["Value"]
        if len(data["Records"]) == 0:
            return S_OK(0)
        pNames = list(data["ParameterNames"])
        dnPos = pNames.index("UserDN")
        groupPos = pNames.index("UserGroup")
        expiryPos = pNames.index("ExpirationTime")
        for row in data["Records"]:
            if DN == row[dnPos] and group == row[groupPos]:
                td = row[expiryPos] - datetime.datetime.utcnow()
                secondsLeft = td.days * 86400 + td.seconds
                return S_OK(max(0, secondsLeft))
        return S_OK(0)

    def getUserProxiesInfo(self):
        """Get the user proxies uploaded info

        :return: S_OK(dict)/S_ERROR()
        """
        result = RPCClient("Framework/ProxyManager", timeout=120).getUserProxiesInfo()
        if "rpcStub" in result:
            result.pop("rpcStub")
        return result


gProxyManager = ProxyManagerClient()



#!/usr/bin/env python

"""
Populates the database with the current installations of components
This script assumes that the InstalledComponentsDB, the
ComponentMonitoring service and the Notification service are installed and running
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


from datetime import datetime
from DIRAC import exit as DIRACexit
from DIRAC import S_OK, gLogger, gConfig
from DIRAC.ConfigurationSystem.Client.CSAPI import CSAPI
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script
from DIRAC.FrameworkSystem.Client.NotificationClient import NotificationClient
from DIRAC.FrameworkSystem.Client.SystemAdministratorIntegrator import SystemAdministratorIntegrator
from DIRAC.FrameworkSystem.Client.ComponentMonitoringClient import ComponentMonitoringClient
from DIRAC.FrameworkSystem.Utilities import MonitoringUtilities
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.Core.Security.ProxyInfo import getProxyInfo

__RCSID__ = "$Id$"

global excludedHosts
excludedHosts = []


def setExcludedHosts(value):
    global excludedHosts

    excludedHosts = value.split(",")
    return S_OK()


@Script()
def main():
    global excludedHosts
    Script.registerSwitch(
        "e:", "exclude=", "Comma separated list of hosts to be excluded from the scanning process", setExcludedHosts
    )
    Script.parseCommandLine(ignoreErrors=False)

    componentType = ""

    # Get my setup
    mySetup = gConfig.getValue("DIRAC/Setup")

    # Retrieve information from all the hosts
    client = SystemAdministratorIntegrator(exclude=excludedHosts)
    resultAll = client.getOverallStatus()
    if not resultAll["OK"]:
        gLogger.error(resultAll["Message"])
        DIRACexit(-1)

    # Retrieve user installing the component
    result = getProxyInfo()
    if result["OK"]:
        user = result["Value"]["username"]
    else:
        DIRACexit(-1)
    if not user:
        user = "unknown"

    for host in resultAll["Value"]:
        if not resultAll["Value"][host]["OK"]:
            # If the host cannot be contacted, exclude it and send message
            excludedHosts.append(host)

            result = NotificationClient().sendMail(
                Operations().getValue("EMail/Production", []),
                "Unreachable host",
                "\ndirac-populate-component-db: Could not fill the database with the components from unreachable host %s\n"
                % host,
            )
            if not result["OK"]:
                gLogger.error("Can not send unreachable host notification mail: %s" % result["Message"])

    resultHosts = client.getHostInfo()
    if not resultHosts["OK"]:
        gLogger.error(resultHosts["Message"])
        DIRACexit(-1)
    resultInfo = client.getInfo()
    if not resultInfo["OK"]:
        gLogger.error(resultInfo["Message"])
        DIRACexit(-1)
    resultMySQL = client.getMySQLStatus()
    if not resultMySQL["OK"]:
        gLogger.error(resultMySQL["Message"])
        DIRACexit(-1)
    resultAllDB = client.getDatabases()
    if not resultAllDB["OK"]:
        gLogger.error(resultAllDB["Message"])
        DIRACexit(-1)
    resultAvailableDB = client.getAvailableDatabases()
    if not resultAvailableDB["OK"]:
        gLogger.error(resultAvailableDB["Message"])
        DIRACexit(-1)

    records = []
    finalSet = list(set(resultAll["Value"]) - set(excludedHosts))
    for host in finalSet:
        hasMySQL = True
        result = resultAll["Value"][host]
        hostResult = resultHosts["Value"][host]
        infoResult = resultInfo["Value"][host]
        mySQLResult = resultMySQL["Value"][host]
        allDBResult = resultAllDB["Value"][host]
        availableDBResult = resultAvailableDB["Value"][host]

        if not result["OK"]:
            gLogger.error("Host %s: %s" % (host, result["Message"]))
            continue
        if not hostResult["OK"]:
            gLogger.error("Host %s: %s" % (host, hostResult["Message"]))
            continue
        if not infoResult["OK"]:
            gLogger.error("Host %s: %s" % (host, infoResult["Message"]))
            continue
        if mySQLResult["OK"]:
            if not allDBResult["OK"]:
                gLogger.error("Host %s: %s" % (host, allDBResult["Message"]))
                continue
            if not availableDBResult["OK"]:
                gLogger.error("Host %s: %s" % (host, availableDBResult["Message"]))
                continue
        else:
            hasMySQL = False

        setup = infoResult["Value"]["Setup"]
        if setup != mySetup:
            continue

        cpu = hostResult["Value"]["CPUModel"].strip()
        rDict = result["Value"]
        # Components other than databases
        for compType in rDict:
            if componentType and componentType != compType:
                continue
            for system in rDict[compType]:
                components = sorted(rDict[compType][system])
                for component in components:
                    record = {"Installation": {}, "Component": {}, "Host": {}}
                    if rDict[compType][system][component]["Installed"] and component != "ComponentMonitoring":
                        runitStatus = str(rDict[compType][system][component]["RunitStatus"])
                        if runitStatus != "Unknown":
                            module = str(rDict[compType][system][component]["Module"])
                            record["Component"]["System"] = system
                            record["Component"]["Module"] = module
                            # Transform 'Services' into 'service', 'Agents' into 'agent' ...
                            record["Component"]["Type"] = compType.lower()[:-1]
                            record["Host"]["HostName"] = host
                            record["Host"]["CPU"] = cpu
                            record["Installation"]["Instance"] = component
                            record["Installation"]["InstallationTime"] = datetime.utcnow()
                            record["Installation"]["InstalledBy"] = user
                            records.append(record)

        # Databases
        csClient = CSAPI()
        cfg = csClient.getCurrentCFG()["Value"]

        if hasMySQL:
            allDB = allDBResult["Value"]
            availableDB = availableDBResult["Value"]

            for db in allDB:
                # Check for DIRAC only databases
                if db in availableDB and db != "InstalledComponentsDB":
                    # Check for 'installed' databases
                    isSection = cfg.isSection(
                        "Systems/"
                        + availableDB[db]["System"]
                        + "/"
                        + cfg.getOption("DIRAC/Setups/" + setup + "/" + availableDB[db]["System"])
                        + "/Databases/"
                        + db
                        + "/"
                    )
                    if isSection:
                        record = {"Installation": {}, "Component": {}, "Host": {}}
                        record["Component"]["System"] = availableDB[db]["System"]
                        record["Component"]["Module"] = db
                        record["Component"]["Type"] = "DB"
                        record["Host"]["HostName"] = host
                        record["Host"]["CPU"] = cpu
                        record["Installation"]["Instance"] = db
                        record["Installation"]["InstallationTime"] = datetime.utcnow()
                        record["Installation"]["InstalledBy"] = user
                        records.append(record)

    monitoringClient = ComponentMonitoringClient()

    # Add the installations to the database
    for record in records:
        result = MonitoringUtilities.monitorInstallation(
            record["Component"]["Type"],
            record["Component"]["System"],
            record["Installation"]["Instance"],
            record["Component"]["Module"],
            record["Host"]["CPU"],
            record["Host"]["HostName"],
        )
        if not result["OK"]:
            gLogger.error(result["Message"])


if __name__ == "__main__":
    main()
